/* FILE: add_builtin.c
                            Peter Lemkin 

                   Image Processing Section, DCBD
                      National Cancer Institute
                        FCRF Bld 469, Rm 150B
                         Frederick, MD 21701

			 Revised/UNIX January 25, 1988
			 June 7, 1987

     COPYRIGHT NOTICE - Copyright (c) 1983-1988 by Peter F. Lemkin.
                        All rights reserved.
	Permission is granted to any individual or institution to
        copy, use or modify this program, except for explicitly
        commercial purposes, provided this copyright notice is
        included in its entirety in all derived works.  No warranty
        of the software or of the accuracy of any documentation
        surrounding it is expressed or implied, and the author does
        not acknowledge any liability resulting from program or
        documentation errors.

              Derived from PSAIL.SAI V.43.50   7-19-84
*/


/* File: psinit.c, being converted 12/25/87, 01:41:42PM*/
/* PSAIL compiler, version: June 9, 1987 V.57.10(OPTIMIZED) 	*/

#include "pshead.h"	/* header file to include other header files */


/*  ----------------------------------------------------- */
/*   --
.LABEL [<++PSAIL PROCEDURE: ADD!BUILTIN>]
.INDEX ++PSAIL PROCEDURE: ADD!BUILTIN
DESCRIPTION:
	Process the symbol-spec string 'symbolSpec' by picking out
the 6 fields/entry and saving them in the hash table entry h (on the
SAIL (1st) symbol. Where h is hash table index and idx is
declXXX[idx] symtab entry index.
	<optional> subLanguage name (Li or other name)
	saiSym[h] - sail token symbol (will cvt to l.c.)
	declCsym[idx] - c token symbol (keep case)
	declFSM[h] - semantic processing for this rule if not zero.
	(Optional type declar. as beginning of 4th arg ==>declDW[idx],
		with remainder of formal args ==> declFormal[idx])
	(Optional # of args of 5th arg in declNargs[idx])
	(optional comment if preceeded by white space)

Note: 	'declNargs[]' is part of symbol table and is the number of args
in a procedure. If the number of args is variable, then it is < 0.

This procedure is called from SET!EXTENSIONS, BUILD!TABLES during
symbol table initialization and when invoked by PSAIL!DEFINE [FSM49].

SIDE-EFFECTS:
	It enters the symbol into the symbol table. If it is there
already, then it will cause a fatal error since it is a DRYROT class
bug. If we are redefining a symbol already in Lk with a new symbol in
Lm where (m>k) then do not consider this an error but let it go ahead
and redefine the symbol.

PROBLEMS/KNOWN-BUGS:

.ENDLABEL [<++PSAIL PROCEDURE: ADD!BUILTIN>]

	List of SAIL operators and runtimes which ARE normally
handled in PSAIL.

The entry is made up of up to 7 fields.
[<Li>:]<SAIname>,<Cname>,<FSM#>[,<decls.>,<#args>]  [</ * comments * />]

Fields within [...] in the above example are optional.

NOTE: if the 5th field is a SAIL declaration type (for a TYPED built
in function), then we also put the symbol into the declSymbol[] with
the type in which case a 5th filed is the # args of the function.
NOTE: @V.45.01 change field 4 to contain optional PROCEDURE type 
specification as follows:
	outType:argList1:n1[arglist2:n2[...,
Where argListI is the Ith declaration list.
The type is given as a SAIL declaration type with 
	array		is coded as	'a'
	string		is coded as	'c'
	double		is coded as (MODIFIED LONG REAL==double) 'd'
	event		is coded as	'e'
	real		is coded as	'f'
	label		is coded as	'g'
	integer		is coded as	'i'
	object		is coded as (NEW PSAIL TYPE)	'j'
	matrix		is coded as (NEW PSAIL TYPE)	'k'
	list		is coded as	'l'
	macro		is coded as 	'm'
	void		is coded as (NEW PSAIL type) 'n'
	optional	is coded as (NEW PSAIL type) 'o'
	process		is coded as (MODIFIED PSAIL type) 'p'
	record_pointer	is coded as	'q'
	reference	is coded as	'r'
	set		is coded as	's'
	record!class	is coded as	't'
	untyped		is coded as (NEW PSAIL TYPE)  'u'
	itemvar		is coded as	'v'
	own		is coded as	'w'
	item		is coded as	'x'
	record!field	is coded as (MODIFIED PSAIL TYPE) 'y'
	safe		is coded as	'z'
**      Boolean         is coded as	'i'
**      itemArray       is coded as	'{'
**      Short           is coded as	'|'
**      Byte           is coded as	'}'
**      C!ARRAY        is coded as	'~'

  For example,
	String Procedure ABC(Integer p,d;
			 Real q; Ref Real x; String s; Integer i,j,k)
	would have the specification:
		c:i:2[f:1[rf:1[c:1[i:3,

 The symbols kept in ASCII files:

   PSMSAI.INI - Symbols for BUILD!SYMTAB: MSAINSAIL symbols /MAINSAIL.
   PSAIL.INI  - Symbols for BUILD!SYMTAB: normally used SAIL oprs/fcts.
   PSEXTN.INI - Symbols for BUILD!SYMTAB: PSAIL /EXTENSION symbols.

   */

void add_builtin(symbolspec)
    STRING symbolspec;
{/* ADD!BUILTIN */
  STRING s, 
	s1, 
	s2, 
	s3, 
	sem, 
	oldsymspec;
  INTEGER h,
	i, 
	ch, 
	brk;

/*  # # DBUG('1,"[ADD!BUILTIN] symbolSpec='"&symbolSpec&"'") */

  GC_POPS(symbolspec); /* from GC string PDL */
  GC_MARKSSTKFRAME;   /* Mark G.C. STACK-FRAME */
  GC_NOTESTRING(symbolspec); /* for G.C. */
  GC_NOTESTRING(s); /* for G.C. */
  GC_NOTESTRING(s1); /* for G.C. */
  GC_NOTESTRING(s2); /* for G.C. */
  GC_NOTESTRING(s3); /* for G.C. */
  GC_NOTESTRING(sem); /* for G.C. */
  GC_NOTESTRING(oldsymspec); /* for G.C. */

  oldsymspec= symbolspec;		/*  -- in case need for err msg */
  if (*symbolspec=='\0')
    { 
      GC_UNMARKSSTKFRAME;
      GC_RETURN(s);
    }
  if (*symbolspec!=':')
    s1= GC_EPOPS(cvlower(scan(&symbolspec,13L,&brk)));
  else s1= GC_EPOPS(cvlower(scan(&symbolspec,2L,&brk)));  /* 1st arg */
  lastdeclaration= 0100L;
  if (brk==':')
    {/* Sub language name specified */
      saitoken= GC_EPOPS(cvlower(scan(&symbolspec,2L,&brk)));  /* 2nd arg */
      if (*s1=='l'&&
	  chk_number(GC_PUSHS(s2= subst(GC_PUSHS(s1),2L,length(s1)))) &&
	  (0<=(lastsublang= cvd(GC_PUSHS(s2))) && lastsublang<=15))
	{/* map the Li number to symbol */
	  /*  - - - DUMMY for now - - - */
	}/* map the Li number to symbol */
      else lastsublang= 0L;
    }/* Sub language name specified */

  else   
    {/* no sublanguage specified, assume <sairun> */
      saitoken= s1;		/*  -- this WAS the symbol */
      lastsublang= 0L;
    }/* no sublanguage specified, assume <sairun> */
  ctoken= GC_EPOPS(scan(&symbolspec,2L,&brk));	/*  -- 2nd arg */
  sem= GC_EPOPS(scan(&symbolspec,2L,&brk));	/*  -- 3rd arg */
  lastfsm= cvd(GC_PUSHS(sem));

  /*  -- @V.56.144 Allow redefinition of symbol in higher value
   *     sublanguage with out error message.
   */
  if (((h= hash_c(GC_PUSHS(saitoken)))!=0) &&
      (lastsublang<=gdw_sublanguage((*aChkI(decldw,1L,
					    (*aChkI(dwptr,1L,h)))))))
    err_hlr(GC_PUSHS(catlist(3L,
		  GC_PUSHS("[ADD!BUILTIN] DRYROT - multiply def'ed token ["),
			     GC_PUSHS(saitoken),GC_PUSHS("]")))); 
  else  
    {/* new token */
      ++nsym;
    }/* new token */

  lastcsym= ctoken;	/*  -- If not new, then redefine it */

  lastnargs= 0L;
  /*  # # DBUG('1,"[ADD!BUILTIN].1.1 saiToken='"&saiToken&
      "' lastFSM='"&CVS(lastFSM)&"' symbolSpec='"&symbolSpec&"'") */
  /*  NOTE: as of @V.48.22 -- put EVERYTHING into permanent symbol table-- */
  /*  -- Note: @V.45.01 the format of the procedure type is:
   *  For example,
   String Procedure ABC(Integer p,d;
			 Real q; Ref Real x; String s; Integer i,j,k)
			 would have the specification:
			 c:i:1[f:1[rf:1[c:1[i:3,
			 See cvTable for specification...
 */
  s2= GC_EPOPS(scan(&symbolspec,2L,&brk)); /* ',' pickup the declar list */
  /*  # # DBUG('1,"[ADD!BUILTIN].1.2 s2='"&s2&"' brk='"&
               (ps!sTrash!_brk)&"' symbolSpec='"&symbolSpec&"' symbolSpec='"&
               symbolSpec&"'") ;
  */
  s3= PNULL;
  if (*s2==':') 
    brk= LOP(s2); 
  else s3= GC_EPOPS(scan(&s2,13L,&brk));   /* return declar ':', can be NULL */
  /*  # # DBUG('1,"[ADD!BUILTIN].2 symbolSpec='"&symbolSpec&"' s2='"&s2&
      "'"&CRLF&" s3='"&s3&"'") */
  if (*s3=='\0'&&lastfsm==71)
    s3= GC_EPOPS(makstr(codevoid));
  /*  # # DBUG('1,"[ADD!BUILTIN].2.0.1 new s3='"&s3&"' lastDeclaration="&
      CVOS(lastDeclaration)&"(8)") */
  if (('a'<=*s3 && *s3<='z') ||*s3=='{' ||*s3=='|' ||*s3=='}' ||
      *s3=='~')
    pdw_ptype(lastdeclaration,(INTEGER)*s3); 
                                         /* Note: s3 is internal PSAIL
					  * letter code! See defns
					  * in cvTable */
  if (!(isalpha(GC_PUSHS(s2)) ||*s2==':'))
    lastformal= PNULL; 
  else lastformal= s2;			/*  -- CVT '[' to ','  */
  if (*symbolspec!='-' && !isdigit((INTEGER)*symbolspec))
    lastnargs= 0L;
  else lastnargs= cvd(GC_PUSHS(symbolspec));
  /*  # # DBUG('1,"[ADD!BUILTIN].2.1 saiToken='"&saiToken&
      "' lastCsym='"&lastCsym&"' "&CRLF&
      " lastDeclaration="&CVOS(lastDeclaration)&
      "(8)='"&CVTYPES(lastDeclaration)&"' lastFormal='"&lastFormal&"'") */
  lastasize= "0";
  if (lastfsm==71||lastfsm==10 ||lastfsm==32 ||lastfsm==33 ||
      (52<=lastfsm && lastfsm<=59)) lastprocedure= "<runtime>"; 
  else lastprocedure= "<main>";
  GC_EPOPS(map_declar(GC_PUSHS(saitoken))); 
                                       /* put LOWER CASE token into tables. */
  lastprocedure= "<main>";
  lastfsm= 0L;	/*  -- clear out whatever else we need cleared */
  if ((h= hash_c(GC_PUSHS(saitoken)))==0) 
    err_hlr(catlist(7L,GC_PUSHS("[ADD!BUILTIN] DRYROT - HASH!C("),
		    GC_PUSHS(saitoken),GC_PUSHS(") is 0."),GC_PUSHS(crlf),
		    GC_PUSHS(" symbolSpec=\'"),GC_PUSHS(oldsymspec),
		    GC_PUSHS("\'")));
  /*  # # DBUG('1,"[ADD!BUILTIN].3 saiToken='"&saiToken&"' h="&CVS(h)&
      " spec='"&oldSymSpec&"'") */

  {/* INCSTAT */
    if (usereportflag)
      (*aChkI(errstatistics,1L,23L))++;
  }/* INCSTAT */				/*  -- for /REPORT */

/*************************************************/
/*        U N U S E D    D E C L A R A T I O N S */
/* 'ch'=<L0> integer                             */
/* 'i'=<L0> integer                              */
/* 's'=<L0> string                               */
/*************************************************/

  {GC_UNMARKSSTKFRAME;}
}/* ADD!BUILTIN */
