/* FILE: build_tables.c
                            Peter Lemkin 

                   Image Processing Section, DCBD
                      National Cancer Institute
                        FCRF Bld 469, Rm 150B
                         Frederick, MD 21701

			 Revised/UNIX January 26, 1988
			    June 7, 1987

     COPYRIGHT NOTICE - Copyright (c) 1983-1988 by Peter F. Lemkin.
                        All rights reserved.
	Permission is granted to any individual or institution to
        copy, use or modify this program, except for explicitly
        commercial purposes, provided this copyright notice is
        included in its entirety in all derived works.  No warranty
        of the software or of the accuracy of any documentation
        surrounding it is expressed or implied, and the author does
        not acknowledge any liability resulting from program or
        documentation errors.

              Derived from PSAIL.SAI V.43.50   7-19-84
*/


/* File: psinit.c, being converted 12/25/87, 01:41:42PM*/
/* PSAIL compiler, version: June 9, 1987 V.57.10(OPTIMIZED) 	*/

#include "pshead.h"	/* header file to include other header files */

/*  --
.LABEL [<++PSAIL PROCEDURE: BUILD!TABLES>]
.INDEX ++PSAIL PROCEDURE: BUILD!TABLES
DESCRIPTION:
	Process the macro String 'cvTable' by picking out the 5
fields/entry and saving them in the hash indexed (on the SAIL (1st)
symbol. See ADD!BUILTIN() for details.

SIDE-EFFECTS:

PROBLEMS/KNOWN-BUGS:

.ENDLABEL [<++PSAIL PROCEDURE: BUILD!TABLES>]
*/

/* 
	List of SAIL operators and runtimes which ARE normally
handled in PSAIL.

The entry is made up of up to 7 fields.
[<Li>:]<SAIname>,<Cname>,<FSM#>[,<decls.>,<#args>]  [</ * comments * />]

Fields within [...] in the above example are optional.

NOTE: if the 5th field is a SAIL declaration type (for a TYPED built
in function), then we also put the symbol into the declSymbol[] with
the type in which case a 5th filed is the # args of the function.
NOTE: @V.45.01 change field 4 to contain optional PROCEDURE type 
specification as follows:
	outType:argList1:n1[arglist2:n2[...,
Where argListI is the Ith declaration list.
The type is given as a SAIL declaration type with 
	array		is coded as	'a'
	string		is coded as	'c'
	double		is coded as (MODIFIED LONG REAL==double) 'd'
	event		is coded as	'e'
	real		is coded as	'f'
	label		is coded as	'g'
	integer		is coded as	'i'
	object		is coded as (NEW PSAIL TYPE)	'j'
	matrix		is coded as (NEW PSAIL TYPE)	'k'
	list		is coded as	'l'
	macro		is coded as 	'm'
	void		is coded as (NEW PSAIL type) 'n'
	optional	is coded as (NEW PSAIL type) 'o'
	process		is coded as (MODIFIED PSAIL type) 'p'
	record_pointer	is coded as	'q'
	reference	is coded as	'r'
	set		is coded as	's'
	record!class	is coded as	't'
	untyped		is coded as (NEW PSAIL TYPE)  'u'
	itemvar		is coded as	'v'
	own		is coded as	'w'
	item		is coded as	'x'
	record!field	is coded as (MODIFIED PSAIL TYPE) 'y'
	safe		is coded as	'z'
**      Boolean         is coded as	'i'
**      itemArray       is coded as	'{'
**      Short           is coded as	'|'
**      Byte           is coded as	'}'
**      C!ARRAY        is coded as	'~'
*/

/*  For example,
	String Procedure ABC(Integer p,d;
			 Real q; Ref Real x; String s; Integer i,j,k)
	would have the specification:
		c:i:2[f:1[rf:1[c:1[i:3,

The symbols kept in ASCII files:

   PSMSAI.INI - Symbols for BUILD!SYMTAB: MSAINSAIL symbols /MAINSAIL.
   PSAIL.INI  - Symbols for BUILD!SYMTAB: normally used SAIL oprs/fcts.
   PSEXTN.INI - Symbols for BUILD!SYMTAB: PSAIL /EXTENSION symbols.
*/

/*  ----------------------------------------------------- */
 void build_tables()

{/* BUILD!TABLES */
  STRING 
    s, 
    s1, 
    sdata;
  INTEGER 
    cs,
    hs, 
    forcedtype;

  GC_MARKSSTKFRAME;   /* Mark G.C. STACK-FRAME */
  GC_NOTESTRING(s); /* for G.C. */
  GC_NOTESTRING(s1); /* for G.C. */
  GC_NOTESTRING(sdata); /* for G.C. */

  GC_EPOPS(dbug(01L,catlist(2L,GC_PUSHS("[BUILD!TABLES] useExtensionsFlag="),
			    pbool(useextensionsflag))));

  pushnumber(debugswitch);	/*  -- Turn it off while building tables */
  debugswitch= 0L;
  nsym= 0L;
  if (*((s= GC_EPOPS(readfile(GC_PUSHS("PSAIL.INI")))))=='?')
    usererr(0L,0L,catlist(2L,GC_PUSHS("[BUILD!TABLES] "),GC_PUSHS(s)));

#if (parsemainsail==1)/
  /* MESSAGE: " <Parsing MAINSAIL code>"  */;
  if (mainsailflag) 
    {/* /MAINSAIL */
      if (*((sdata= GC_EPOPS(readfile(GC_PUSHS("PSMSAI.INI")))))=='?')
	PSprintf("%s%s",GC_PUSHS(sdata),GC_PUSHS(crlf)); 
      else s= catlist(2L,GC_PUSHS(s),GC_PUSHS(sdata));
    }/* /MAINSAIL */
#endif  

  while ((*s!='\0'))
    {/* build table */
      s1= GC_EPOPS(scan(&s,1L,&saibrk)); /*next line*/
      pushnumber(debugswitch);
      debugswitch= 1L;
      /*  # # DBUG('1,"[BUILD!TABLES].1 s1='"&s1&"'") */
      popnumber(&debugswitch);
      if ((*s1=='\0')) 
	continue;
      add_builtin(GC_PUSHS(s1));	      /*  -- to permanent symtab */
    }/* build table */

  /*  --	.  +  .  +  .  +  .  +  .  +  .  +  .  +   */
  /*  --	NOTE: add special extra-tabular symbols here.... */
  if (nbrcmtflag)
    add_builtin(GC_PUSHS("L0:#,comment,10,"));	/*  -- /# flag */

  lastcsym= ",";		/*  -- since can't put ',' in cvTABLE!!! */
  lastsublang= 0L;		/*  -- Sublanguage L0 */
  lastfsm= 0L;		/*  -- simple map */
  lastformal= PNULL;
  lastnargs= 0L;
  lastasize= "0";
  lastdeclaration= 0100L;
  GC_EPOPS(map_declar(GC_PUSHS(",")));

  lastcsym= "<main>";
  GC_EPOPS(map_declar(GC_PUSHS(lastcsym)));

  lastcsym= "<runtime>";
  GC_EPOPS(map_declar(GC_PUSHS(lastcsym)));
  nsym += 3;

  /* If /CRLF then use "\n" instead */
  if (usecrlfflag)
    (*aChkC(declcsym,1L,find_symbol(GC_PUSHS("crlf"),&cs,&hs))) = 
      catlist(GC_PUSHS(quote),GC_PUSHS("\\n"),
	      GC_PUSHS(quote));

  lastsublang= 1L;		/*  -- Sublanguage L1 */
  forcedtype= 0;
  pdw_ptype(forcedtype,coderecclass);
  (*aChkI(decldw,1L,
	  find_symbol(GC_PUSHS("any_class"),&cs,&hs)))= forcedtype;

  forcedtype= 020000L;
  pdw_ptype(forcedtype,codeitem);
  pdw_itemtype(forcedtype,1L);	/*  -- untyped item */
  (*aChkI(decldw,1L,
	  find_symbol(GC_PUSHS("bindit"),&cs,&hs)))= forcedtype;

  /* If /REAL:double then reset it */
  if (!userealflag)
    {/* change to DOUBLE */
      INTEGER idx;
      idx= find_symbol(GC_PUSHS("real"),&cs,&hs);
      pdw_ptype((*aChkI(decldw,1L,idx)),codedouble);	/*  -- Redefine */
      (*aChkC(declcsym,1L,idx))= "double";
    }/* change to DOUBLE */


  /* Set up hash index of SAIL Built in key words ---- */
  hany_class = hash_c(GC_PUSHS("any_class"));
  hand = hash_c(GC_PUSHS("and"));
  hash = hash_c(GC_PUSHS("ash"));
  hassoc = hash_c(GC_PUSHS("assoc"));
  hbegin = hash_c(GC_PUSHS("begin"));
  hboolean = hash_c(GC_PUSHS("boolean"));
  hcase = hash_c(GC_PUSHS("case"));
  hcomment = hash_c(GC_PUSHS("comment"));
  hc_code = hash_c(GC_PUSHS("c_code"));
  hdefine = hash_c(GC_PUSHS("define"));
  hdiv = hash_c(GC_PUSHS("div"));
  hdo = hash_c(GC_PUSHS("do"));
  helse = hash_c(GC_PUSHS("else"));
  helsec = hash_c(GC_PUSHS("elsec"));
  hend = hash_c(GC_PUSHS("end"));
  hendc = hash_c(GC_PUSHS("endc"));
  hevaldefine = hash_c(GC_PUSHS("evaldefine"));
  hevalinclude = hash_c(GC_PUSHS("evalinclude"));
  hevalredefine = hash_c(GC_PUSHS("evalredefine"));
  hexternal = hash_c(GC_PUSHS("external"));
  hfor = hash_c(GC_PUSHS("for"));
  hforeach = hash_c(GC_PUSHS("foreach"));
  hforward = hash_c(GC_PUSHS("forward"));
  hgeq = hash_c(GC_PUSHS("geq"));
  hif = hash_c(GC_PUSHS("if"));
  hifc = hash_c(GC_PUSHS("ifc"));
  hin = hash_c(GC_PUSHS("in"));
  hinter = hash_c(GC_PUSHS("inter"));
  hinteger = hash_c(GC_PUSHS("integer"));
  hintrnal = hash_c(GC_PUSHS("internal"));
  hitem = hash_c(GC_PUSHS("item"));
  hitemvar = hash_c(GC_PUSHS("itemvar"));
  hlabel = hash_c(GC_PUSHS("label"));
  hland = hash_c(GC_PUSHS("land"));
  hleap_array = hash_c(GC_PUSHS("leap_array"));
  hlength = hash_c(GC_PUSHS("length"));
  hleq = hash_c(GC_PUSHS("leq"));
  hlist = hash_c(GC_PUSHS("list"));
  hlistc = hash_c(GC_PUSHS("listc"));
  hlisto = hash_c(GC_PUSHS("listo"));
  hlong = hash_c(GC_PUSHS("long"));
  hlnot = hash_c(GC_PUSHS("lnot"));
  hlop = hash_c(GC_PUSHS("lop"));
  hlor = hash_c(GC_PUSHS("lor"));
  hlsh = hash_c(GC_PUSHS("lsh"));
  hmax = hash_c(GC_PUSHS("max"));
  hmin = hash_c(GC_PUSHS("min"));
  hmod = hash_c(GC_PUSHS("mod"));
  hneq = hash_c(GC_PUSHS("neq"));
  hnomac = hash_c(GC_PUSHS("nomac"));
  hnot = hash_c(GC_PUSHS("not"));
  hor = hash_c(GC_PUSHS("or"));
  hprint = hash_c(GC_PUSHS("print"));
  hprocedure = hash_c(GC_PUSHS("procedure"));
  hquick_code = hash_c(GC_PUSHS("quick_code"));
  hreal = hash_c(GC_PUSHS("real"));
  hrecursive = hash_c(GC_PUSHS("recursive"));
  hredefine = hash_c(GC_PUSHS("redefine"));
  hreference = hash_c(GC_PUSHS("reference"));
  href_item = hash_c(GC_PUSHS("ref_item"));
  hremove = hash_c(GC_PUSHS("remove"));
  hrot = hash_c(GC_PUSHS("rot"));
  hsafe = hash_c(GC_PUSHS("safe"));
  hset = hash_c(GC_PUSHS("set"));
  hsetc = hash_c(GC_PUSHS("setc"));
  hseto = hash_c(GC_PUSHS("seto"));
  hsimple = hash_c(GC_PUSHS("simple"));
  hstart_code = hash_c(GC_PUSHS("start_code"));
  hstep = hash_c(GC_PUSHS("step"));
  hstring = hash_c(GC_PUSHS("string"));
  hthenc = hash_c(GC_PUSHS("thenc"));
  hunion= hash_c(GC_PUSHS("union"));
  huntil = hash_c(GC_PUSHS("until"));
  hvalue = hash_c(GC_PUSHS("value"));
  hwhile = hash_c(GC_PUSHS("while"));
  hxor = hash_c(GC_PUSHS("xor"));

  hbackarrow = hash_c(GC_PUSHS(":="));
  hcolon= hash_c(GC_PUSHS(":"));
  hdivide_sign = hash_c(GC_PUSHS("/"));
  hleftbracket = hash_c(GC_PUSHS("["));
  hleftparen = hash_c(GC_PUSHS("("));
  hleftsquiggle = hash_c(GC_PUSHS("{"));
  hminus = hash_c(GC_PUSHS("-"));
  hquote = hash_c(GC_PUSHS(quote));
  hsemicolon = hash_c(GC_PUSHS(";"));

  h_main_ = hash_c(GC_PUSHS("<MAIN>"));
  h_runtime_ = hash_c(GC_PUSHS("<runtime>"));

  set_extensions(useextensionsflag); /*  -- set up L3 symbols
					 when /EXTENSIONS and undefine
					 PSAIL L3 extensions when
					 /NOEXTENSIONS 
					 */

#if (parsemainsail==1)
  /*  -- N.B. If you really want to add MAINSAIL parsing capability, then
      we need to extensively edit this.
      */
  hbits = hash_c(GC_PUSHS("bits"));
  haddress = hash_c(GC_PUSHS("address"));
  hcharadr = hash_c(GC_PUSHS("charadr"));
  hclass = hash_c(GC_PUSHS("class"));
  hmsk = hash_c(GC_PUSHS("msk"));
  hior = hash_c(GC_PUSHS("ior"));
  hclr = hash_c(GC_PUSHS("clr"));
  hshl = hash_c(GC_PUSHS("shl"));
  hshr = hash_c(GC_PUSHS("shr"));
  htst = hash_c(GC_PUSHS("tst"));
  hntst = hash_c(GC_PUSHS("ntst"));
  htsta = hash_c(GC_PUSHS("tsta"));
  hntsta = hash_c(GC_PUSHS("ntsta"));
  hprint = hash_c(GC_PUSHS("ttywrite")); /*  -- @V.56.46 special fakeout */
#endif
  
  (*aChkI(dcllocstkframe,1L,0L))= dclsymptr;  
                                /* save for possible use by /UNIQUE */

  popnumber(&debugswitch);	/* Turn it back on if active */

  {GC_UNMARKSSTKFRAME;}
}/* BUILD!TABLES */
