/* FILE: chk_c_keyword.c
                            Peter Lemkin 

                   Image Processing Section, DCBD
                      National Cancer Institute
                        FCRF Bld 469, Rm 150B
                         Frederick, MD 21701

			  April 7, 1987

     COPYRIGHT NOTICE - Copyright (c) 1983-1988 by Peter F. Lemkin.
                        All rights reserved.
	Permission is granted to any individual or institution to
        copy, use or modify this program, except for explicitly
        commercial purposes, provided this copyright notice is
        included in its entirety in all derived works.  No warranty
        of the software or of the accuracy of any documentation
        surrounding it is expressed or implied, and the author does
        not acknowledge any liability resulting from program or
        documentation errors.

              Derived from PSAIL.SAI V.43.50   7-19-84
 */


/* File: pschek.c, being converted 12/25/87, 02:41:52PM*/
/* PSAIL compiler, version: June 9, 1987 V.57.10(OPTIMIZED) 	*/

#include "pshead.h"	/* header file to include other header files */

/*  --
.LABEL [<++PSAIL PROCEDURE: CHK!C!KEYWORD>]
.INDEX ++PSAIL PROCEDURE: CHK!C!KEYWORD
DESCRIPTION:
	If token, which is assumed to be a lower case token, is a C
keyword (as defined by membership in global preloaded array
cKeywords[1:maxCkeywords]), then return TRUE else FALSE. The test
uses a hash table lookup algorithm such that the selection of the
hashing function and hash table size to insure that there are NO
collisions so we merely have to compute the initial hash probe and
test if anything is there. If there is then we have a possible
mapping conflict which is tested further with an actual string
comparison.

SIDE-EFFECTS:

PROBLEMS/KNOWN-BUGS:

.ENDLABEL [<++PSAIL PROCEDURE: CHK!C!KEYWORD>]
 */

/*  ----------------------------------------------------- */
INTEGER chk_c_keyword(token)
	STRING token;

{/* CHK!C!KEYWORD */
  INTEGER 
    val, 
    i, 
    h, 
    h0;
  STRING 
    s, 
    s1, 
    s2;
  static INTEGER 
    setupflag = 0;		/*  -- will have PI if was here before */
#define maxchash 178 		/*  -- smallest prime # > *maxCkeywords
				  -- such that no clashes occur. NOTE
				  -- that no UPPERCASE keywords are put
				  -- into table */
  static INTEGER *chashtbl= NULL;  /* [0:maxchash] */ 
                                   /* hash table of ptrs
				    * to C keywords in Ckeywords[] */
#define chashfct(stok,hh) {hh= 0; s1= stok;\
	while (*s1!='\0') \
		hh= 02 + LOP(s1);   /*  -- simple hash fct */\
	h0= hh= 01L;\
	}



  /*  -------------------------------------------- */
  /*  -- [0]	Init hash table if first time */
  /*  -------------------------------------------- */
  GC_POPS(token); /* from GC string PDL */
  GC_MARKSSTKFRAME;   /* Mark G.C. STACK-FRAME */
  GC_NOTESTRING(token); /* for G.C. */
  GC_NOTESTRING(s); /* for G.C. */
  GC_NOTESTRING(s1); /* for G.C. */
  GC_NOTESTRING(s2); /* for G.C. */
  chashtbl = arymkI(1L,0L,178L,"chashtbl",'i');	/* Dynamic Array */
  if (setupflag!=314159)
    {/* init C keyword hash table for rapid lookups */
      setupflag= 314159L;
      arrclr(chashtbl,0L);
      for (i=1;i<=40L;i++)
	{/* put C keyword into hash table */
	  s2= (*aChkC(ckeywords,1L,i));
	  if ('A'<=*s2 && *s2<='Z')
	    continue;		/* UPPER CASE ARE OK so do
				 * not put into hash table */
	  chashfct((INTEGER)*s2,h);
	  if ((*aChkI(chashtbl,1L,h))!=0)
	    err_hlr(GC_PUSHS(
			     catlist(11L,
				     GC_PUSHS(
  "[CHK!C!KEYWORD] DRYROT - cHashTbl[0:maxChash=178] is not unique."),
				     GC_PUSHS(crlf), 
				     GC_PUSHS("cKeywords[cHashTbl["), cvs(h),
				     GC_PUSHS("]=\'"), 
				     GC_PUSHS(*aChkC(ckeywords,1L,
						     (*aChkI(chashtbl,1L,h)))),
				     GC_PUSHS("\' and s2=\'"),
				     GC_PUSHS(s2),GC_PUSHS("\'"),
				     GC_PUSHS(crlf),
			     GC_PUSHS("Pick another maxChash table size.")))); 
	  else (*aChkI(chashtbl,1L,h))= i;   /*  -- set index to C name */
	}/* put C keyword into hash table */
    }/* init C keyword hash table for rapid lookups */


  /*  -------------------------------------------- */
  /*  -- [1] Go look up the token and map if match */
  /*  -------------------------------------------- */
  s= token;
  chashfct((INTEGER)*token,h);		/*  -- lookup hash index if any */
  val= ((i= (*aChkI(chashtbl,1L,h)))!=0) && 
    equ(cvlower(GC_PUSHS((*aChkC(ckeywords,1L,i)))),GC_PUSHS(token)) &&
      (*token!='\0');		/* If hash index was not 0 then see if
				 * exact match of token with C keyword */
  GC_EPOPS(dbug(02000000L,catlist(4L,GC_PUSHS("[CHK!C!KEYWORD] s=\'"),
				  GC_PUSHS(s),
				  GC_PUSHS("\' val="),pbool(val))));


  /*  ---------------------------------------------------------- */
  /*  -- [2] If Not C keyword name conflict and using /NAMEMAPing then
      check if the token is in the namemap table. */
/*  ---------------------------------------------------------- */
  if ((!val) &&usenamemapflag) 
    {/* see if need to do namemap */
      s= GC_EPOPS(chk_name_map(GC_PUSHS(token)));	/* Iff /NAMEMAP */
      val= fast_equ(GC_PUSHS(token),GC_PUSHS(s));
    }/* see if need to do namemap */
  GC_EPOPS(dbug(02000000L,catlist(5L,GC_PUSHS("[CHK!C!KEYWORD].2 val="),
				  pbool(val),
				  GC_PUSHS(" s=\'"),
				  GC_PUSHS(s),GC_PUSHS("\'"))));

  GC_NRETURN(val)
#undef chashfct
#undef maxchash
}/* CHK!C!KEYWORD */
