/* FILE: copy_either.c
                            Peter Lemkin 

                   Image Processing Section, DCBD
                      National Cancer Institute
                        FCRF Bld 469, Rm 150B
                         Frederick, MD 21701

			    June 7, 1987

     COPYRIGHT NOTICE - Copyright (c) 1983-1988 by Peter F. Lemkin.
                        All rights reserved.
	Permission is granted to any individual or institution to
        copy, use or modify this program, except for explicitly
        commercial purposes, provided this copyright notice is
        included in its entirety in all derived works.  No warranty
        of the software or of the accuracy of any documentation
        surrounding it is expressed or implied, and the author does
        not acknowledge any liability resulting from program or
        documentation errors.

              Derived from PSAIL.SAI V.43.50   7-19-84
 */


/* File: pslex.c, being converted 12/25/87, 08:38:10PM*/
/* PSAIL compiler, version: June 9, 1987 V.57.10(OPTIMIZED) 	*/

#include "pshead.h"	/* header file to include other header files */


/*  ------------------------------------------------------- */
/*INTERNAL*/ STRING copy_either(symlist,brkstr)
	STRING symlist;
	STRING *brkstr;
/*  ------------------------------------------------------- */
{/* COPY!EITHER */
/*   --
.LABEL [<++PSAIL PROCEDURE: COPY!EITHER>]
.INDEX ++PSAIL PROCEDURE: COPY!EITHER
DESCRIPTION:
	COPY!EITHER is similar in intent to SCAN(). It is alexical
token scanner for break symbols (rather than just characters).

	Scan from global variable 'saiStatement' (or input file if at
R!D level 0). Return all characters up to but not including the
(symbols) given in the 'symList' variable which is a string of a list
of symbols to break on separated by ','s. However, symList may also
include a ',' as the first element of brkStr. The 'saiStatement'
string contains whatever is left after the break if it occurs. The
break matched string (if any) is put in 'brkStr'. NOTE: only symbols
in the HASH table may be checked. So if you want to check a new
symbol make sure we put it in the HASH table in BUILD!TABLE...

Note that 'rcLevel' is the recursion level of COPY!EITHER using its
own state stacking routines PUSH!LEVEL and POP!LEVEL.

SIDE-EFFECTS:
	@V.47.91 N.B. - NEVER NEVER allow macro expansion (via
GET!TOKEN) from COPY!EITHER since EXPAND!MACRO() will call
COPY!EITHER and violate the state of COPY!EITHER!!!!!

PROBLEMS/KNOWN-BUGS:
	@V.53.39 Added MAGIC code to handle nested IFC...ENDC syntax
by counting up ifcLevel with IFC and down with ENDC. All of the other
keywords are ignored except when ifcLevel=0. In addition, we have to
prep the ifcLevel to 1 if the string does not have an IFC but does
have ENDC or ELSEC. This is done by pushing a dummy BELL&'ifc' in the
front of the string. This is detected and ifcLevel set to -314159.
The IFC sub-FSM here traps it and sets ifcLevel to 1 as desired. This
cludge string prefix is currently only set in [FSM61] (i.e. IFC
processor) and in [RIGHTSTATEMENT].

	@V.53.91 - moved PUSH!LEVEL macros (called 5 times) to
commonPUSHLEVEL label and use DO!PUSH!LEVEL(x) macros to set up the
name x in a temp variable and then 'goto' commonPUSHLEVEL.

	@V.53.92 added more <goto>'s because the <while-do>"check"
block was too big to parse with PSAIL as it cause STRG GC ovf - so it
was changed to an <if> <goto>scanLOOP/failedMATCH cludge .... sorry
Nick...

NOTE: @V.53.91 NEW version After optimizations of 8-3-85 @V.53.90 --.

.ENDLABEL [<++PSAIL PROCEDURE: COPY!EITHER>]
 */


/*  -- Since we are using a few GOTO's (the 4 letter word!) to speed
things up, it would be nice to be able to pick them out of this
forest ... so all labels are surrounded by '# +++++++' as for example:
		# +++++++++++
		# +++++++++++
		sampleLABEL:
		# +++++++++++
		# +++++++++++
 */
/* LABEL  chknullinput */		/*  -- instead of CONTINUE... when
				  -- we are done processing the
				  -- current token */
/* LABEL  commonpushlevel */		/*  -- common entry for DO!PUSH!LEVEL */
/* LABEL  failedmatch */		/*  -- failed the MATCH */
/* LABEL  scanloop */			/*  -- Scan loop to check tokens */
/* LABEL  testmatch */		/*  -- if YOU KNOW WHO YOU ARE...
				  -- and you want to test a match */


STRING	s, 
	s1, 
	s2, 
	s3		/*  -- temp variables */,
	str			/*  -- string being constructed
				  -- which is returned if succeed */,
	strpushleveltmp	/*  -- set by DO!PUSH!LEVEL(x) */,
	searchlist		/*  -- current break token search list */,
	token;			/*  -- current token being analyzed */


INTEGER
	blevel			/*  -- local working block level for 
				  --  use by COPY!EITHER when scanning
				  --  BEGIN/END type pairs. */,
	bottomlevel		/*  -- block level where start search */,
	brk			/*  -- SCAN() break character */,
	brktcnt		/*  -- '[]' nesting count */,
	ch			/*  -- temp character */,
	chkbits		/*  -- enab/srch for non-lex hash codes */,
	code			/*  -- ... */,
	firstbrkspec		/*  -- used if COMMA is 1st break char */,
	i			/*  -- temp variable */,
	idx			/*  -- symtab index for current token */,
	ifclevel		/*  -- IfC nesting level */,
	h			/*  -- hash code of current token */,
	lasthtrigger		/*  -- hash code of token triggering the 
				  --  last PUSH!LEVEL */,
	lokforbits		/*  -- search for break hash lex codes */,
	marknpdl		/*  -- copy of nPDLptr at start */,
	markspdl		/*  -- copy of sPDLptr at start */,
	match			/*  -- flag set TRUE if match */,
	n			/*  -- number of non-special break
				  -- tokens to check. */,
	nlevels		/*  -- Current PUSH!LEVEL nesting */,
	parencnt		/*  -- '()' nesting count */,
	rclevel		/*  -- recursion level of COPY!EITHER */,
	sqglcnt		/*  -- '{}' nesting count */,
	startblocklevel	/*  -- block level where start search
				  -- for current rcLevel */,
	toklexbit;		/*  -- lex bit found w/hash code lookup */

INTEGER onceflag		/*  -- needed to get us into the
				"check" look at least once incase the
				saiStatement is NULL and we need to
				call GET!NEXT!STATEMENT there... */,
	sawstepflag;		/*  -- flag set by STEP, cleared by UNTIL
				  -- or DO and tested for by WHILE 
				  -- checks */
 static INTEGER beenherebefore = 0;	/*  -- used to init HASH table lex codes */

 STRING /*ARRAY*/ *brklist  /* [0:10] */ ;   /*  -- current BREAK strings */

#define maxlevels 30	    /*  -- maximum number of nested searches */
 STRING /*ARRAY*/ *schlist  /* [0:maxlevels] */ ;   /*  -- searchList @ nLevels */


/*  -- NOTE: we will store these bits in LH (32-bits) of declFSM[] and
compositions of these bits are used in lokForBits and chkBits.
 */

#define bdelim 0400000
#define bfor 01000000
#define bdo 02000000
#define bwhile 04000000
#define bend 010000000
#define bendc 020000000
#define brbrkt 040000000
#define brparen 0100000000
#define brsquiggle 0200000000
#define bsemicolon 0400000000
#define bcomma 01000000000
#define bjustchk 020000000000


#define maxlexcodes 19

 static INTEGER /*ARRAY*/ P_bcodes[13+(19-1+1L)] = { 
   0,0,0,0,0,0,0,0,0,0,0,0,0, bfor,
   	bdo,
   	bwhile,
   	bend,
   	bendc,
   	brbrkt,
   	brparen,
   	brsquiggle,
   	bsemicolon,
   	bcomma,
   		/*  -- @V.50.01 !!! check usage - may need to 
			     add to ths FSM !!! */
   
   	bjustchk,	/*  -- FOREACH */
   	bjustchk,	/*  -- IF */
   	bjustchk,	/*  -- IFC */
   	bjustchk,	/*  -- BEGIN */
   	bjustchk,	/*  -- START!CODE */
   	bjustchk,	/*  -- QUICK!CODE */
   	bjustchk,	/*  -- C!CODE */
   	bjustchk,	/*  -- STEP */
   	bjustchk };
 static INTEGER /*ARRAY*/ *bcodes  /* [1:maxlexcodes] */ ;


/*  -- the following token list contains tokens which might be
  -- break strings and those which will never be break strings but
  -- might modify the FSM search */
 static STRING /*ARRAY*/ 
/******************************************************************/
/* WARNING! Can't have OWN STRING declarations in non-MAIN module */
/* or inner blocks! Suggest you move 'STRING scodes' declaration  */
/* to global level in MAIN module and recompile.                  */
/******************************************************************//*N.P.*/
/***************************/
/*<7>    ..                */
/*      .                  */
/*   /----------\          */
/*   | o         _______/  */
/*   |__         -------\  */
/*   \----------/          */
/*                         */
/*   A GROSS MOBY          */
/***************************/

P_scodes[26+(19-1+1L)] = { 
   0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, "for",
   	"do",
   	"while",
   	"end",
   	"endc",
   	"]",
   	")",
   	"}",
   	";",
   	",",
   
   	"foreach",
       /*  -- We will NEVER specify these as break strings */
   	"if",
   	"ifc",
   	"begin",
   	"start_code",
   	"quick_code",
   	"c_code",
   	"step",
   	"until" };
 static STRING /*ARRAY*/ 
/******************************************************************/
/* WARNING! Can't have OWN STRING declarations in non-MAIN module */
/* or inner blocks! Suggest you move 'STRING scodes' declaration  */
/* to global level in MAIN module and recompile.                  */
/******************************************************************//*N.P.*/
/***************************/
/*<7>    ..                */
/*      .                  */
/*   /----------\          */
/*   | o         _______/  */
/*   |__         -------\  */
/*   \----------/          */
/*                         */
/*   A GROSS MOBY          */
/***************************/

*scodes  /* [1:maxlexcodes] */ ;



/*  ------------------------------------------------------------- */
/*  [0] . . LEVEL STACKING/UNSTACKING BREAK STRING MACROS . . .  */
/*  ------------------------------------------------------------- */
#define plvl (catlist(3L,GC_PUSHS("{"),cvs(rclevel),\
        GC_PUSHS((usetimerflag) ?  catlist(2L,run_time(),GC_PUSHS(\
        "}")): "}")))


#define chk_fix_pdls {"CHK!FIX!PDLS" \
/*  _2 _2 dbug(010,plvl&"[CHK!FIX!PDLS] sPDLptr="&cvs(spdlptr)&" markSPDL="&cvs(markspdl)&" nPDLptr="&cvs(npdlptr)&" markNPDL="&cvs(marknpdl)) */\
			if ((spdlptr!=markspdl))\
				{"Restore PSAIL sPDL[] frame"\
				/*  -- _this is in case we got out of sync \
				  because of a failure */\
				for (i=(markspdl+1L);i<=spdlptr;i++) (*aChkC(spdl,1L,i))= PNULL;  /*  -- in case gc */\
				spdlptr= markspdl;\
				}/* Restore PSAIL sPDL[] frame */\
			if ((npdlptr!=marknpdl))\
				{"Restore PSAIL nPDL[] frame"\
				/*  -- _this is in case we got out of sync \
				  because of a failure */\
				npdlptr= marknpdl;\
				}/* Restore PSAIL nPDL[] frame */\
			}/* CHK!FIX!PDLS */ 


/* #define set_up_break(blist) begin "SET!UP!BREAK"\
		n:=0;   _2 -- number of non-special _break tokens to check;\
		firstbrkspec:=true;  _2 -- to handle 0,0 special case;\
		s:=blist;	_2 -- list of _break strings to check;\
		lokforbits:=0;	_2 -- _default is to check none;\
		while (s neq null) do\
			begin "build break token list"\
			if (n geq 10)\
				then err_hlr(\
				"[COPY!EITHER] > 10 search symbols!"&\
							crlf&blist);\
			if firstbrkspec and s=","\
				then s1:=lop(s)\
			else s1:=scan(s,2,brk); _2 -- brk on 0,0;\
			firstbrkspec:=false;\
			h:=hash_c(s1);\
			_2 -- note: store b-codes in lh of 32-bit word;\
			toklexbit:=declfsm[dwptr[h]] land 037777600000;\
			if (toklexbit land 017777600000) neq 0\
			    then lokforbits:=lokforbits lor toklexbit\
			else if s1=sailbodydelim or s1=sailargdelim\
			    then lokforbits:=lokforbits lor bdelim\
			else brklist[n:=n+1]:=s1;  _2 -- non-std _break;\
_2 _2 _2 dbug(010,plvl&"[SET!UP!BREAK] bListArg='"&blist&"' brkList["&cvs(n)&"]='"&brklist[n]&"' s1='"&s1&"'"&crlf&"  lokForBits=0"&cvos(lokforbits)&" tokLEXbit=0"&cvos(toklexbit));\
			end "build break token list";\
		lokforbits:=(017777600000 land lokforbits); _2 -- not\
						just check bit;\
		chkbits:=(037777600000 xor lokforbits) lor \
				(bjustchk lor bend lor bendc);\
						_2 -- only check what we\
						are not looking for.\
						always check any\
						justchk tokens;\
		startblocklevel:=blevel;\
_2 _2 _2 dbug(010,plvl&"[SET!UP!BREAK] lokForBits="&cvos(lokforbits)&" chkBits="&cvos(chkbits)&" n="&cvos(n));\
		end "SET!UP!BREAK" ====> PSAIL!EVAL */


/* #define push_level(x) begin "PUSH!LEVEL"\
		_2 -- note: the new _break string was preset into\
			0strpushleveltmp0 and passed here as 0x0;\
		pushnumber(lasthtrigger);\
		lasthtrigger:=h;		_2 -- i.e. remember who did it_;\
		schlist[nlevels]:=searchlist;    _2 -- in case error;\
		if (nlevels:=nlevels+1) > maxlevels\
			then err_hlr(\
"[COPY!EITHER] DRYROT - PUSH!LEVEL overflow 'nLevels>'"&cvs(maxlevels));\
		push(searchlist);\
		searchlist:=x;		_2 -- use new search list;\
		push(str);\
		pushnumber(startblocklevel);	_2 -- a number_;\
		startblocklevel:=blevel;\
		for i:=1 step 1 until n do push(brklist[i]);\
		pushnumber(n);		_2 -- save number;\
		pushnumber(chkbits);\
		pushnumber(lokforbits);\
		rclevel:=rclevel+1;\
		set_up_break(x);	_2 -- set up new _break word;\
dbug(010,plvl&"[PUSH!LEVEL] startBlockLevel="&cvs(startblocklevel)&" nLevels="&cvs(nlevels)&crlf&" lastHtrigger='"&saisym[lasthtrigger]&"' searchList='"&searchlist&"'");\
		str:=null;\
		end "PUSH!LEVEL" ====> PSAIL!EVAL */


#define do_push_level(x) {"DO!PUSH!LEVEL"\
			strpushleveltmp= x;	 /*  -- save it here\
						 until we get there... */\
			goto commonpushlevel;   /*  -- avoid dupl code.. */\
			}/* DO!PUSH!LEVEL */


#define pop_level --{"POP!LEVEL"\
		nlevels;\
		--rclevel;\
		popnumber(&lokforbits);\
		popnumber(&chkbits);\
		popnumber(&n);			/*  -- restore number */\
		for (i=n;i<=1;i--) pop(aChkC(brklist,1L,i));\
		popnumber(&startblocklevel);\
		pop(&s);				/*  -- old str */\
		str= catlist(2L,GC_PUSHS(s),GC_PUSHS(str));			/*  -- append it */\
		pop(&searchlist);\
		popnumber(&lasthtrigger);\
GC_EPOPS(dbug(010L,catlist(10L,GC_PUSHS("&[POP!LEVEL] startBlockLevel="),\
        cvs(startblocklevel),GC_PUSHS(" nLevels="),cvs(nlevels),\
        GC_PUSHS(crlf),GC_PUSHS(" lastHtrigger=\'"),\
        GC_PUSHS(*aChkC(saisym,1L,lasthtrigger)),\
        GC_PUSHS("\' searchList=\'"),GC_PUSHS(searchlist),\
        GC_PUSHS("\'"))));\
		}/* POP!LEVEL */

/*  . . . . . . . . . . END OF MACROS . . . . . . . . . . . . . . .  */


;

/*  ------------------------------- */
/*  [1] . . INITIALIZE SEARCH . . .  */
/*  ------------------------------- */
GC_EPOPS(dbug(010L,catlist(10L,GC_PUSHS("&[COPY!EITHER] symList=\'"),GC_PUSHS(symlist),
        GC_PUSHS("\'"),GC_PUSHS(crlf),GC_PUSHS("  saiStatement=\'"),
        GC_PUSHS(saistatement),GC_PUSHS("\'"),GC_PUSHS(crlf),
        GC_PUSHS("  isFRAGMENTflag="),pbool(isfragmentflag))));
GC_POPS(symlist); /* from GC string PDL */
GC_MARKSSTKFRAME;   /* Mark G.C. STACK-FRAME */
GC_NOTESTRING(symlist); /* for G.C. */
GC_NOTESTRING(brkstr); /* for G.C. */
GC_NOTESTRING(s); /* for G.C. */
GC_NOTESTRING(s1); /* for G.C. */
GC_NOTESTRING(s2); /* for G.C. */
GC_NOTESTRING(s3); /* for G.C. */
GC_NOTESTRING(str); /* for G.C. */
GC_NOTESTRING(strpushleveltmp); /* for G.C. */
GC_NOTESTRING(searchlist); /* for G.C. */
GC_NOTESTRING(token); /* for G.C. */
GC_NOTEASTRING(brklist);	/* for G.C. */
GC_NOTEASTRING(schlist);	/* for G.C. */
GC_NOTEASTRING(scodes);	/* for G.C. */
brklist = arymkC(1L,0L,10L,"brklist",'c');	/* Dynamic Array */
schlist = arymkC(1L,0L,30L,"schlist",'c');	/* Dynamic Array */
bcodes = aryHdrI(P_bcodes,1,1,19,"bcodes",'i');	/* Dynamic Array */
scodes = aryHdrC(P_scodes,1,1,19,"scodes",'c');	/* Dynamic Array */
if (beenherebefore!=314159)
	{/* Initialize hashcode lex codes */
	for (i=1;i<=19L;i++)
		{/* insert code into LEX field of declFSM[] LH */
		idx= (*aChkI(dwptr,1L,hash_c(GC_PUSHS((*aChkC(scodes,1L,i))))));
		(*aChkI(declfsm,1L,idx))= ((*aChkI(declfsm,1L,idx))&0177777)|(*aChkI(bcodes,1L,i));
/*  # # DBUG('10,PLVL&"[COPY!EITHER].0.0.1 sCodes["&CVS(i)&"]='"&sCodes[i]&"' bCodes=0"&CVOS(bCodes[i])&" declFSM=0"&CVOS(declFSM[idx])) */
		}/* insert code into LEX field of declFSM[] LH */
	beenherebefore= 314159L;	  	/*  -- set so we are here... */
	}/* Initialize hashcode lex codes */

pushnumber(iscopy_toflag); iscopy_toflag= 1L;
brktcnt= 0;				/*  -- [ ] counter */
if ((*saistatement==007) &&equ(subst(GC_PUSHS(saistatement),2L,4L),GC_PUSHS("ifc")))
	{/* set ifcLevel to MAGIC and eat 'BELL&ifc' */
	ifclevel= -314159;		/*  -- special MAGIC value */
	brk= LOP(saistatement);		/*  -- first eat the BELL */
	}/* set ifcLevel to MAGIC and eat 'BELL&ifc' */
else ifclevel= 0;	/*  -- 'IFC... ENDC' nesting level */

lasthtrigger= 0;		/*  -- token hash code triggering PUSH!LEVEL */
markspdl= spdlptr;	/*  -- Mark PSAIL string pdl used with
					PUSH()/POP() in case need
					to abort with fragment error! */
marknpdl= npdlptr;	/*  -- Mark PSAIL number pdl */
nlevels= 0;		/*  -- current PUSH!LEVEL nesting  */
onceflag= 1L;		/*  -- Flag to let use get into loop ONCE! This is
			  in case saiStatement is NULL and we
			  have to get into the loop in order to
			  be able to call GET!NEXT!STATEMENT there */
parencnt= 0;		/*  -- (  )  counter */
rclevel= 0;		/*  -- recursion level of COPY!EITHER */
sawstepflag= 0L;	/*  -- will be set if we see STEP to abort 
			     the WHILE---DO test */
sqglcnt= 0;		/*  -- {  } counter */
str= PNULL;		/*  -- Output string */
startblocklevel= bottomlevel= blevel= blocklevel;  /*  -- actual block level */

searchlist= symlist;	/*  -- Initialize break list */
{/* SET!UP!BREAK */
		n= 0;   /*  -- Number of non-special break tokens to check */
		firstbrkspec= 1L;  /*  -- to handle ',' special case */
		s= searchlist;	/*  -- list of break strings to check */
		lokforbits= 0;	/*  -- default is to check NONE */
		while ((*s!='\0'))
			{/* build break token list */
			if ((n>=10)) err_hlr(GC_PUSHS(
				
				        							catlist(GC_PUSHS(
				        "[COPY!EITHER] > 10 search symbols!"),GC_PUSHS(crlf),
				        GC_PUSHS(searchlist))));
			if (firstbrkspec &&*s==',') s1= GC_EPOPS(makstr(LOP(s)
			)); else s1= GC_EPOPS(scan(&s,2L,&brk)); /*  -- brk on ',' */
			firstbrkspec= 0L;
			h= hash_c(GC_PUSHS(s1));
			/*  -- NOTE: store b-Codes in LH of 32-bit word */
			toklexbit= (*aChkI(declfsm,1L,(*aChkI(dwptr,1L,h))))&037777600000;
			if ((toklexbit&017777600000)!=0) lokforbits=lokforbits|toklexbit; 
			else if (*s1==*sailbodydelim||*s1==*sailargdelim) lokforbits=lokforbits|0400000L; 
			else (*aChkC(brklist,1L,++n))= s1;  /*  -- non-STD break */
/*  # # DBUG('10,PLVL&"[SET!UP!BREAK] bListArg='"&searchList&"' brkList["&CVS(n)&"]='"&brkList[n]&"' s1='"&s1&"'"&CRLF&"  lokForBits=0"&CVOS(lokForBits)&" tokLEXbit=0"&CVOS(tokLEXbit)) */
			}/* build break token list */
		lokforbits= (017777600000&lokforbits); /*  -- Not
						JUST CHECK bit */
		chkbits= (037777600000^lokforbits)|0225125746700;
						/*  -- only check what we
						are not looking for.
						Always check any
						JUSTCHK tokens */
		startblocklevel= blevel;
/*  # # DBUG('10,PLVL&"[SET!UP!BREAK] lokForBits="&CVOS(lokForBits)&" chkBits="&CVOS(chkBits)&" n="&CVOS(n)) */
		}/* SET!UP!BREAK */	/*  -- Initialize break list */
/*  # # DBUG('10,PLVL&"[COPY!EITHER].0.0.2 startBlockLevel="&CVS(startBlockLevel)&" n="&CVS(n)&" searchList='"&searchList&"'"&CRLF&" saiStatement='"&saiStatement&"'"&CRLF&" sPDLptr="&CVS(sPDLptr)&" nPDLptr="&CVS(nPDLptr)) */


/*  ---------------------------------------------------------------- */
/*  [2] . . SCANNING LOOP - LOOKING FOR BREAK STRING AT LEVEL 0 . . . */
/*  ---------------------------------------------------------------- */
/*  ++++++++ */
/*  ++++++++ */
scanloop:
/*  ++++++++ */
/*  ++++++++ */
if (!(((*saistatement!='\0') &&(!isfragmentflag)) ||onceflag)) goto failedmatch;
/*  # # @V.53.91 	was <while-do> Begin "check"   */
	onceflag= 0L;		/*  -- only get here for free ONCE! */
/*  # # NOTE: @V.46.73 some S.O.B. put NULLs in the input stream! Find
the bastard! */
	token= GC_EPOPS(get_token(&saistatement));    /*  -- REMEMBER - NO MACRO
						EXPANSION!!!! This
						is because we set
						isCOPY!TOflag true
						here. */
	*brkstr= PNULL;			/*  -- Force to be NULL */
	toklexbit= h= 0;		 	/*  -- Start afresh */
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1 token='"&token&"' parenCnt="&CVS(parenCnt)&" brktCnt="&CVS(brktCnt)&" sqglCnt="&CVS(sqglcnt)&CRLF&" str='"&str&"'"&CRLF&" saiStatement='"&saiStatement&"'"&CRLF&" sPDLptr="&CVS(sPDLptr)&" nPDLptr="&CVS(nPDLptr)&CRLF&" bLevel="&CVS(bLevel)&" ifcLevel="&CVS(ifcLevel)) */
if (rdlevel<1) /* NO-OP */0L

/*  -------------------------------------------------------- */
/*  [2.1]  . CHECK FOR SPECIAL DELIMITERS '(), [], {}' . . . */
/*  -------------------------------------------------------- */
	ch= (INTEGER)*token;		/*  -- make integer for speed */
	if (!(isalpha(makstr(ch)) ||isdigit(ch)))
	 	{/* test brkts */
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.2 NOT-ALPHA token='"&token&"'") */
		if ((ch!=*quote) &&
		   (ch==*sailbodydelim||ch==*sailargdelim) &&
		   (chkbits&0400000L))
			{/* Copy non-quote macro body */
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.2.1 copy DELIMITED string - token='"&token&"' "&CRLF&" str='"&str&"'"&CRLF&" saiStatement='"&saiStatement&"'") */
			if (ch==*sailbodydelim) b24= sairbodydelim; 
			else b24= sairargdelim;
			setbreak(24L,GC_PUSHS(b24),GC_PUSHS(PNULL),GC_PUSHS("INA"));  /*  -- KEEP brkChr */
			str= catlist(3L,GC_PUSHS(str),GC_PUSHS(token),scan(&saistatement,24L,&brk));
			if (brk=='\0')
				err_hlr(GC_PUSHS(
catlist(7L,
        GC_PUSHS("[COPY!EITHER] DRYROT - missing RIGHT-MACRO delimiter \'")
        ,GC_PUSHS(b24),GC_PUSHS("\'"),GC_PUSHS(crlf),
        GC_PUSHS("  Scanned string=\'"),GC_PUSHS(str),GC_PUSHS("\'"))));
			token= PNULL;
			toklexbit= 0400000L;
			}/* Copy non-quote macro body */
		else if (ch=='(') ++parencnt; else if (ch==')') 
			{/* close ')' */
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.2.2 parenCnt="&CVS(parenCnt)) */
			if ((--parencnt)<0) 
				{/* test if BREAK */
				toklexbit= 0100000000L;
				goto testmatch;
				}/* test if BREAK */
			str= catlist(2L,GC_PUSHS(str),GC_PUSHS(token));
			token= PNULL;
			goto chknullinput;
			}/* close ')' */
		else if (ch=='[') ++brktcnt; else if (ch==']') 
			{/* close ']' */
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.2.3 brktCnt="&CVS(brktCnt)) */
			if ((--brktcnt)<0) 
				{/* test if BREAK */
				toklexbit= 040000000L;
				goto testmatch;
				}/* test if BREAK */
			str= catlist(2L,GC_PUSHS(str),GC_PUSHS(token));
			token= PNULL;
			goto chknullinput;
			}/* close ']' */
		else if (ch=='{'&&(ch!=*sailbodydelim) &&
		        (ch!=*sailargdelim)) ++sqglcnt; else if (ch=='}'&&(ch!=*sailbodydelim) &&
			(ch!=*sailargdelim)) 
			{/* close '}' */
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.2.4 sqglCnt="&CVS(sqglcnt)) */
			if ((--sqglcnt)<0) 
				{/* test if BREAK */
				toklexbit= 0200000000L;
				goto testmatch;
				}/* test if BREAK */
			str= catlist(2L,GC_PUSHS(str),GC_PUSHS(token));
			token= PNULL;
			goto chknullinput;
			}/* close '}' */

		   else if (h==hdivide_sign&&*saistatement=='*') 
			{/* C / *...* / */
			INTEGER clvl, 
			ch;
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.2.5 / *...* /") */
			str= catlist(2L,GC_PUSHS(str),GC_PUSHS(token));
			clvl= 0;		/*  -- assume NO COMMENT */
			if ((*saistatement=='*') &&(clvl= 1))
				{/* Map /* to COMMENT */
				str= catlist(2L,GC_PUSHS(str),GC_PUSHS(LOP(saistatement)));
				while ((ch= LOP(saistatement)!='\0'))
					if ((ch=='*'&&
					    *saistatement=='/') ||
					   (ch=='/'&&
					    *saistatement=='*'))
						{/* fnd cmt end/begin */
						s= catlist(3L,GC_PUSHS(s),makstr(ch),GC_PUSHS(LOP(saistatement)));
						if (ch=='/') ++clvl; else --clvl;
						if (clvl<=0) break; /*  -- OK */
						}/* fnd cmt end/begin */
					else s= catlist(2L,GC_PUSHS(s),makstr(ch));  /*  -- just copy */
				}/* Map /* to COMMENT */
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.2.5.1 C / *...* / cLvl="&CVS(cLvl)&CRLF&" str='"&str&"'"&CRLF&" saiStatement='"&saiStatement&"'") */
			goto chknullinput;
			}/* C / *...* / */
		}/* test brkts */
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.3 bLevel='"&CVS(bLevel)&"' token='"&token&"'"&CRLF&" saiStatement='"&saiStatement&"'"&CRLF&" str='"&str&"'") */


/*  ------------------------------------------------- */
/*  [2.2]  . TAKE CARE OF WHITESPACE        . . .     */
/*  ------------------------------------------------- */
	if ((*token!='\0') &&
	   iswhitespace((INTEGER)*token) ||(brktcnt>0) ||(parencnt>0) ||
	   (sqglcnt>0))
		{/* just copy W.S. */
		if (*token==*singlequote&&
			(*(saistatement+2-1))==*singlequote)
			{/* handle 'x' */
			str= catlist(4L,GC_PUSHS(str),GC_PUSHS(token),GC_PUSHS(LOP(saistatement)),
        GC_PUSHS(LOP(saistatement)));
			token= GC_EPOPS(makstr(LOP(saistatement)));	/*  -- trash it */
			}/* handle 'x' */
		else if (*token==*quote)
			{/* Copy quoted string to avoid trouble */
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.4 copy quoted string - token='"&token&"' "&CRLF&" str='"&str&"'"&CRLF&" saiStatement='"&saiStatement&"'") */
			str= catlist(2L,GC_PUSHS(str),map_quotes(&saistatement));
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.5 copy quoted string - str='"&str&"'"&CRLF&" saiStatement='"&saiStatement&"'") */
			if (isfragmentflag) goto failedmatch;   /*  -- ABORT!!!!! */
			}/* Copy quoted string to avoid trouble */
			else str= catlist(2L,GC_PUSHS(str),GC_PUSHS(token));
		token= PNULL;
		}/* just copy W.S. */


/*  ----------------------------------------------------------- */
/*  [2.3] . . TAKE CARE OF NON-NULL TOKENS . . .		      */
/*  ----------------------------------------------------------- */
     else if ((*token!='\0'))
	{/* test if token is KEYWORD */
	h= hash_c(GC_PUSHS(token));
	toklexbit= (*aChkI(declfsm,1L,(*aChkI(dwptr,1L,h))))&037777600000;
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.6 bLevel="&CVS(bLevel)&" h="&CVS(h)&" token='"&token&"'"&CRLF&"  tokLEXbit=0"&CVOS(tokLEXbit)&" chkBits=0"&CVOS(chkBits)&" lokForBits=0"&CVOS(lokForBits)&CRLF&" str='"&str&"'"&CRLF&" ifcLevel="&CVS(ifcLevel)) */


/*  ------------------------------------------------ */
/*  [2.3.1] . . CHECK FOR QUOTES AND COMMENTS . . .  */
/*  ------------------------------------------------ */
	if (h!=0)
	   if (h==hcomment) 
		{/* Comment */
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.6.1 COMMENT") */
		str= catlist(4L,GC_PUSHS(str),GC_PUSHS(token),scan(&saistatement,10L,&brk),
        GC_PUSHS(";"));
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.6.1.1 COMMENT str='"&str&"'"&CRLF&" saiStatement="&saiStatement) */
		goto chknullinput;
		}/* Comment */

	else if (h==hquote)
		{/* QUOTE */
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.6.2 QUOTE") */
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.6.2.1 copy QUOTEd string - token='"&token&"' "&CRLF&" str='"&str&"'"&CRLF&" saiStatement='"&saiStatement&"'") */
		str= catlist(2L,GC_PUSHS(str),map_quotes(&saistatement));
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.6.2.2 str='"&str&"' LENGTH(str)="&CVS(LENGTH(str))) */
		if (isfragmentflag) goto failedmatch;		/*  -- ABORT!!!! */
		goto chknullinput;
		}/* QUOTE */


/*  ----------------------------------------------------------------
 *//*  [2.3.2] . . CHECK SPECIAL LEXICAL TOKENS WHICH ARE ACTIVE . . .  */
/*  ---------------------------------------------------------------- */
	else if ((chkbits&toklexbit))
	{/* special LEX code tests */
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.6.3 CHECK SPECIAL LEX-TOKENS sawSTEPflag="&PBOOL(sawSTEPflag)) */


/*  ------------------------------------------------------------ */
/*  [2.3.3] . . USE SPECIAL PROCESSING FOR IFC/ENDC BLOCKS . . . */
/*  ------------------------------------------------------------ */
	if ((ifclevel==0))
	   {/* NOT in IFC...ENDC - check other keywords */
	   if ((h==hbegin||h==hstart_code ||
	        h==hquick_code ||h==hc_code))
		{/* BEGIN token */
		str= catlist(2L,GC_PUSHS(str),GC_PUSHS(token));
		++blevel;
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.6.4 after BEGIN bLevel="&CVS(bLevel)) */
		goto chknullinput;
		}/* BEGIN token */

	else if (h==hend&&(!(lokforbits&010000000L) ||
			    ((lokforbits&010000000L) &&
			     blevel>startblocklevel)))
		{/* END token */
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.6.4.1 before END bLevel="&CVS(bLevel)) */
		if ((--blevel)<0) goto failedmatch;     /*  -- FATAL exit --- */
		if (!(lokforbits&010000000L)) 
			{str= catlist(2L,GC_PUSHS(str),GC_PUSHS(token));
			goto chknullinput;
			}
		}/* END token */

	else if ((h==hwhile&&(chkbits&04000000L)) &&
		(lasthtrigger!=hfor)
				/*   -- NOTE: lastHtrigger for AMBIGUOUS
				-- 'while' which can be a statement in
				-- its own right or can share the 'do'
				-- in a 'for...do' loop --- */)
		{/* WHILE token */
		str= catlist(2L,GC_PUSHS(str),GC_PUSHS(token));
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.6.5 WHILE sawSTEPflag="&PBOOL(sawSTEPflag)&" token='"&token&"'") */
		if (!sawstepflag) GC_EPOPS(do_push_level((INTEGER)('d'/*WARNING*/)));
		goto chknullinput;
		}/* WHILE token */

	else if (((h==hfor&&(chkbits&040000000L) &&(brktcnt==0) &&
		 !isarrayadrflag) ||h==hforeach))
		{/* FOR(each) token */
		str= catlist(2L,GC_PUSHS(str),GC_PUSHS(token));
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.6.6 FOR/FOREACH sawSTEPflag="&PBOOL(sawSTEPflag)) */
		GC_EPOPS(do_push_level((INTEGER)('d'/*WARNING*/)));
		goto chknullinput;
		}/* FOR(each) token */

	else if (h==hdo&&(chkbits&02000000L))
		{/* DO token */
		sawstepflag= 0L;
		str= catlist(2L,GC_PUSHS(str),GC_PUSHS(token));
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.6.7 DO sawSTEPflag="&PBOOL(sawSTEPflag)) */
		GC_EPOPS(do_push_level((INTEGER)('u'/*WARNING*/)));
		goto chknullinput;
		}/* DO token */

	else if ((h==hstep||h==huntil))
		{/* STEP token */
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.6.8 STEP/UNTIL sawSTEPflag="&PBOOL(sawSTEPflag)) */
		sawstepflag= (h==hstep);
		toklexbit= 0;	/*  -- just treat as regular token */
		goto testmatch;	 /*  -- ok, set flag now see if break */
		}/* STEP token */

	else if (h==hif)
		{/* IF token */
		str= catlist(2L,GC_PUSHS(str),GC_PUSHS(token));
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.6.9 IF") */
		GC_EPOPS(do_push_level((INTEGER)('t'/*WARNING*/)));
		goto chknullinput;
		}/* IF token */
	   }/* NOT in IFC...ENDC - check other keywords */

	if (h==hifc)
		{/* IFC token */
		if (ifclevel==-314159) ifclevel= 
		1L; else 	{/* real IFC - stack it */
			str= catlist(2L,GC_PUSHS(str),GC_PUSHS(token));
			++ifclevel;
			GC_EPOPS(do_push_level((INTEGER)('e'/*WARNING*/)));
			}/* real IFC - stack it */
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.6.10 after IFC ifcLevel="&CVS(ifcLevel)&" nLevels="&CVS(nLevels)&CRLF&" str='"&str&"'") */
		goto chknullinput;
		}/* IFC token */

	else if (h==hendc)
		{/* ENDC token */
		--ifclevel;
/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.6.11 after ENDC ifcLevel="&CVS(ifcLevel)) */
		if ((blevel<0) ||(ifclevel<0)) goto failedmatch;     /*  -- FATAL exit --- */
		if (!(lokforbits&020000000L)) 
			{/* just add it on */
			str= catlist(2L,GC_PUSHS(str),GC_PUSHS(token));
			goto chknullinput;
			}/* just add it on */
		else goto testmatch;
		}/* ENDC token */
	}/* special LEX code tests */
	}/* test if token is KEYWORD */

/*  # # DBUG('10,PLVL&"[COPY!EITHER].1.7 token='"&token&"' tokLEXbit=0"&CVOS(tokLEXbit)&CRLF&"  bLevel='"&CVS(bLevel)&"' str='"&str&"'"&CRLF&" LENGTH(str)="&CVS(LENGTH(str))) */


/*  ------------------------------------------------------------------ */
/*  [2.4] . TEST IF TOKEN IS A NON-SPECIAL BREAK STRING w/BALANCE. .   */
/*  ------------------------------------------------------------------ */
			/*  -- ONLY CHECK if at right block level! */
/*  ++++++++++ */
/*  ++++++++++ */
testmatch:		/*  -- this is a real live GOTO.... label... */
/*  ++++++++++ */
/*  ++++++++++ */
	match= (lokforbits&toklexbit)!=0;
/*  # # DBUG('10,PLVL&"[COPY!EITHER].2 token='"&token&"' startBlockLevel="&CVS(startBlockLevel)&" match="&PBOOL(match)&CRLF&"  lokForBits=0"&CVOS(lokForBits)&"  tokLEXbit=0"&CVOS(tokLEXbit)&"  bLevel="&CVS(bLevel)&"  rcLevel="&CVS(rcLevel)) */
	if ((startblocklevel>=blevel) &&
	   ((*token!='\0') ||match)) 
	  {/* Check if found token */
/*  # # DBUG('10,PLVL&"[COPY!EITHER].2.1 match="&PBOOL(match)&" n="&CVS(n)) */
	  if (!match &&n>0) for (i=1;i<=n;i++)
		     if (*((s3= (*aChkC(brklist,1L,i))))==*token&&equ(GC_PUSHS(s3),GC_PUSHS(token)))	{/* found a match */
				match= 1L;
				break;
				}/* found a match */
	  if (match)
		{/* found break token */
		if (*token!='\0') *brkstr= token;  /*  -- save if not special */
GC_EPOPS(dbug(010L,catlist(15L,GC_PUSHS("&[COPY!EITHER].2.2 MATCHED brkStr=\'"),
        GC_PUSHS(*brkstr),GC_PUSHS("\' |brkStr|="),
        cvs(length(GC_PUSHS(*brkstr))),GC_PUSHS(crlf),
        GC_PUSHS("  bottomLevel="),cvs(bottomlevel),
        GC_PUSHS(" bLevel="),cvs(blevel),GC_PUSHS(" nLevels="),
        cvs(nlevels),GC_PUSHS(crlf),GC_PUSHS("  str=\'"),GC_PUSHS(
        str),GC_PUSHS("\'"))));
		if (bottomlevel==blevel&&nlevels==0) 
			{/* OK, send back string and break */
/* NO-OP */;
/* NO-OP */;
			chk_fix_pdls;   /*  -- fixup sPDL[] & nPDL[]
					since may be been left with
					crap in the stacks when aborted
					the search */
			popnumber(&iscopy_toflag);
GC_EPOPS(dbug(010L,catlist(7L,GC_PUSHS("&[COPY!EITHER].2.3 RETURNING brkStr=\'"),
        GC_PUSHS(*brkstr),GC_PUSHS("\'"),GC_PUSHS(crlf),
        GC_PUSHS(" str=\'"),GC_PUSHS(str),GC_PUSHS("\'"))));
;
			{aryfree((INTEGER *)brklist,(INTEGER *)schlist,
			        (INTEGER *)0);  /* Free List of Dynamic Arrays */
			GC_ADJSTKFRAME(2);
			 GC_SRETURN(str);}	/*  -- All finished! */
			}/* OK, send back string and break */


/*  ------------------------------------------------------------------ */
/*  [2.5] . . FOUND BREAK @ INNER LVL - BACKUP LVL & CONT. SEARCH . .  */
/*  ------------------------------------------------------------------ */
		if ((nlevels>0)) 
		    {/* backup a level */
		    pop_level  /*  -- Restore previous search if any */
		    sawstepflag= 0L;	/*  -- reset at each level */
		    if (*token!='\0') str= catlist(2L,GC_PUSHS(str),GC_PUSHS(token));  /*  -- get it if never did */
GC_EPOPS(dbug(010L,catlist(17L,GC_PUSHS("&[COPY!EITHER].2.4 AFTER POP!LEVEL brkStr=\'"),
        GC_PUSHS(*brkstr),GC_PUSHS("\' "),GC_PUSHS(crlf),
        GC_PUSHS("  bottomLevel="),cvs(bottomlevel),
        GC_PUSHS(" bLevel="),cvs(blevel),GC_PUSHS(" nLevels="),
        cvs(nlevels),GC_PUSHS(crlf),GC_PUSHS("  str=\'"),GC_PUSHS(
        str),GC_PUSHS("\'"),GC_PUSHS("  saiStatement=\'"),
        GC_PUSHS(saistatement),GC_PUSHS("\'"))));
		    goto chknullinput;
		    }/* backup a level */

		else if ((bottomlevel>=blevel) ||(nlevels>0)) 
			{/* Error! */
			s= " ";
			for (i=nlevels;i<=1;i--)
				s= catlist(8L,GC_PUSHS(s),GC_PUSHS("at nLevels="),cvs(i),
        GC_PUSHS(" searchList=\'"),GC_PUSHS(*aChkC(schlist,1L,i)),
        GC_PUSHS("\'"),GC_PUSHS(crlf),GC_PUSHS(" "));
			err_hlr(GC_PUSHS(
		catlist(29L,GC_PUSHS("[COPY!EITHER] can\'t find \'"),
		        GC_PUSHS(searchlist),
		        GC_PUSHS("\' AT BOTTOM LEVEL for nLevels="),cvs(nlevels),
		        GC_PUSHS(":"),GC_PUSHS(crlf),GC_PUSHS(s),
		        GC_PUSHS("Start block-level="),cvs(startblocklevel),
		        GC_PUSHS(crlf),GC_PUSHS(" Bottom block-level="),
		        cvs(bottomlevel),GC_PUSHS(crlf),GC_PUSHS(" Block-level="),
		        cvs(blevel),GC_PUSHS(crlf),GC_PUSHS(" IFC-level="),
		        cvs(ifclevel),GC_PUSHS(crlf),GC_PUSHS(" LEX-RECUR-level=")
		        ,cvs(rclevel),GC_PUSHS(crlf),GC_PUSHS(" str=\'"),
		        GC_PUSHS(str),GC_PUSHS("\'"),GC_PUSHS(crlf),
		        GC_PUSHS(" saiStatement=\'"),GC_PUSHS(saistatement),
		        GC_PUSHS("\'"))));
			goto failedmatch;		/*  -- ABORT !!!! */
			}/* Error! */
/*  NOTE: @V.45.138 what about BEGIN s0, s1, ..., IF be THEN s1 END ???
		 */goto failedmatch;
		}/* found break token */
	  }/* Check if found token */

	if (*token!='\0') str= catlist(2L,GC_PUSHS(str),GC_PUSHS(token));  	/*  -- Not found, build string */
GC_EPOPS(dbug(010L,catlist(3L,GC_PUSHS("&[COPY!EITHER].3 NO Match str=\'"),GC_PUSHS(str),
        GC_PUSHS("\'"))));
/*  -- @V.50.12 NOTE: We COULD return the special string '<NOMATCH>'? */


/*  ---------------------------------------------------------------- */
/*  [2.6] . . CHECK FOR NULL INPUT AND READ SOME MORE IF ENABLED . . . */
/*  ---------------------------------------------------------------- */
/*  . . . . . . . . . . . . . . . . . . . . . . . . . . . . .  */
goto chknullinput;	/*  -- skip PAST this HIDDEN spagetti code */
/*  ++++++++++++++++ */
/*  ++++++++++++++++ */
commonpushlevel:
/*  ++++++++++++++++ */
/*  ++++++++++++++++ */
			/*  -- all DO!PUSH!LEVEL(x) end up here
			  -- after setting 'strPushLeveltmp' to 'x' */
{/* PUSH!LEVEL */
		/*  -- NOTE: the new break string was preset into
			'strPushLevelTmp' and passed here as 'strPushLevelTmp' */
		pushnumber(lasthtrigger);
		lasthtrigger= h;		/*  -- i.e. remember who did it! */
		(*aChkC(schlist,1L,nlevels))= searchlist;    /*  -- in case error */
		if ((++nlevels)>30L) err_hlr(GC_PUSHS(
"[COPY!EITHER] DRYROT - PUSH!LEVEL overflow \'nLevels>\'30"));
		push(GC_PUSHS(searchlist));
		searchlist= strpushleveltmp;		/*  -- use new search list */
		push(GC_PUSHS(str));
		pushnumber(startblocklevel);	/*  -- a Number! */
		startblocklevel= blevel;
		for (i=1;i<=n;i++) push(GC_PUSHS((*aChkC(brklist,1L,i))));
		pushnumber(n);		/*  -- Save number */
		pushnumber(chkbits);
		pushnumber(lokforbits);
		++rclevel;
		{/* SET!UP!BREAK */
		n= 0;   /*  -- Number of non-special break tokens to check */
		firstbrkspec= 1L;  /*  -- to handle ',' special case */
		s= strpushleveltmp;	/*  -- list of break strings to check */
		lokforbits= 0;	/*  -- default is to check NONE */
		while ((*s!='\0'))
			{/* build break token list */
			if ((n>=10)) err_hlr(GC_PUSHS(
				
				        							catlist(GC_PUSHS(
				        "[COPY!EITHER] > 10 search symbols!"),GC_PUSHS(crlf),
				        GC_PUSHS(strpushleveltmp))));
			if (firstbrkspec &&*s==',') s1= GC_EPOPS(makstr(LOP(s)
			)); else s1= GC_EPOPS(scan(&s,2L,&brk)); /*  -- brk on ',' */
			firstbrkspec= 0L;
			h= hash_c(GC_PUSHS(s1));
			/*  -- NOTE: store b-Codes in LH of 32-bit word */
			toklexbit= (*aChkI(declfsm,1L,(*aChkI(dwptr,1L,h))))&037777600000;
			if ((toklexbit&017777600000)!=0) lokforbits=lokforbits|toklexbit; 
			else if (*s1==*sailbodydelim||*s1==*sailargdelim) lokforbits=lokforbits|0400000L; 
			else (*aChkC(brklist,1L,++n))= s1;  /*  -- non-STD break */
/*  # # DBUG('10,PLVL&"[SET!UP!BREAK] bListArg='"&strPushLevelTmp&"' brkList["&CVS(n)&"]='"&brkList[n]&"' s1='"&s1&"'"&CRLF&"  lokForBits=0"&CVOS(lokForBits)&" tokLEXbit=0"&CVOS(tokLEXbit)) */
			}/* build break token list */
		lokforbits= (017777600000&lokforbits); /*  -- Not
						JUST CHECK bit */
		chkbits= (037777600000^lokforbits)|0225125746700;
						/*  -- only check what we
						are not looking for.
						Always check any
						JUSTCHK tokens */
		startblocklevel= blevel;
/*  # # DBUG('10,PLVL&"[SET!UP!BREAK] lokForBits="&CVOS(lokForBits)&" chkBits="&CVOS(chkBits)&" n="&CVOS(n)) */
		}/* SET!UP!BREAK */	/*  -- Set up new break word */
GC_EPOPS(dbug(010L,catlist(10L,GC_PUSHS("&[PUSH!LEVEL] startBlockLevel="),
        cvs(startblocklevel),GC_PUSHS(" nLevels="),cvs(nlevels),
        GC_PUSHS(crlf),GC_PUSHS(" lastHtrigger=\'"),
        GC_PUSHS(*aChkC(saisym,1L,lasthtrigger)),
        GC_PUSHS("\' searchList=\'"),GC_PUSHS(searchlist),
        GC_PUSHS("\'"))));
		str= PNULL;
		}/* PUSH!LEVEL */	/*  -- so we expand it only once! */
goto chknullinput;	/*  -- continue processing */
/*  . . . . . . . . . . . . . . . . . . . . . . . . . . . . .  */


/*  ++++++++++++ */
/*  ++++++++++++ */
chknullinput:
/*  ++++++++++++ */
/*  ++++++++++++ */
	if ((*saistatement=='\0') &&(!ismacroflag)) 
		{/* ran out of input, get more! */
/*  # # DBUG('10,PLVL&"[COPY!EITHER].3.1 NUL.INPUT str='"&str&"'"&CRLF&" line='"&line&"'"&CRLF&" saiStatement='"&saiStatement&"'"&CRLF& "At nLevel=" & CVS(nLevels)&" startBlockLevel="&CVS(startBlockLevel)&CRLF&" bottomBlockLevel="&CVS(bottomLevel)&CRLF&" bLevel="&CVS(bLevel)&CRLF&" nPDLptr="&CVS(nPDLptr)&" sPDLptr="&CVS(nPDLptr)) */
		if (rdlevel<1)
			{/* refill statement buffer */
/*  -- NOTE: @V.47.117 We must NEVER push all this stuff since
 --  COPY!EITHER might be called recursively as follows:
 --		1. GET!NEXT!STATEMENT calls CVEVALQUOTE.
 --		2. A token in CVEVALQUOTE may be a macro fragment in 
 --		   which case it calls EXPAND!MACRO.
 --		3. EXPAND!MACRO may need to find actual args and will
 --		   call COPY!EITHER with the match string ",)" and
 --		   this is where we can get into trouble!
 */
GC_EPOPS(dbug(010L,catlist(27L,GC_PUSHS("&[COPY!EITHER].3.1.1 NUL.INPUT str=\'"),
        GC_PUSHS(str),GC_PUSHS("\'"),GC_PUSHS(crlf),
        GC_PUSHS(" line=\'"),GC_PUSHS(line),GC_PUSHS("\'"),
        GC_PUSHS(crlf),GC_PUSHS(" saiStatement=\'"),
        GC_PUSHS(saistatement),GC_PUSHS("\'"),GC_PUSHS(crlf),
        GC_PUSHS("At nLevel="),cvs(nlevels),
        GC_PUSHS(" startBlockLevel="),cvs(startblocklevel),
        GC_PUSHS(crlf),GC_PUSHS(" bottomBlockLevel="),cvs(
        bottomlevel),GC_PUSHS(crlf),GC_PUSHS(" bLevel="),cvs(blevel)
        ,GC_PUSHS(crlf),GC_PUSHS(" nPDLptr="),cvs(npdlptr),
        GC_PUSHS(" sPDLptr="),cvs(npdlptr))));
			iscopy_toflag= 0L;  /*  -- so CVEVALQUOTE works */
			get_next_statement();
			iscopy_toflag= 1L; 	/*  -- back on here */
GC_EPOPS(dbug(010L,catlist(27L,GC_PUSHS("&[COPY!EITHER].3.2 NUL.INPUTstr=\'"),GC_PUSHS(
        str),GC_PUSHS("\'"),GC_PUSHS(crlf),GC_PUSHS(" line=\'"),
        GC_PUSHS(line),GC_PUSHS("\'"),GC_PUSHS(crlf),
        GC_PUSHS(" saiStatement=\'"),GC_PUSHS(saistatement),
        GC_PUSHS("\'"),GC_PUSHS(crlf),GC_PUSHS("At nLevel="),
        cvs(nlevels),GC_PUSHS(" startBlockLevel="),
        cvs(startblocklevel),GC_PUSHS(crlf),
        GC_PUSHS(" bottomBlockLevel="),cvs(bottomlevel),GC_PUSHS(
        crlf),GC_PUSHS(" bLevel="),cvs(blevel),GC_PUSHS(crlf),
        GC_PUSHS(" nPDLptr="),cvs(npdlptr),GC_PUSHS(" sPDLptr="),
        cvs(npdlptr))));
			}/* refill statement buffer */
		}/* ran out of input, get more! */
	goto scanloop;		/*  -- try next token if any */
/*  # # @V.53.91	was <while-do> End "check" */


/*  ---------------------------------------------------------------- */
/*  [3.0] . . FAILED TO FIND MATCH. RESTORE STACKS AND RETURN . . .  */
/*  ---------------------------------------------------------------- */
/*  ++++++++++ */
/*  ++++++++++ */
failedmatch:
/*  ++++++++++ */
/*  ++++++++++ */
chk_fix_pdls;				/*  -- fixup sPDL[] & nPDL[]
					since may be been left with
					crap in the stacks when aborted
					the search */
if (ismacroflag)	{/* special-handling for MACROS */
		/*  -- @V.54.45 the following lets us handle embedded
		     assignments and other possibly legal fragments 
		     within macros ... WE WILL SEE ...
		     Eg.	Define PDQ={x:=(y LSH -3)}
		 */
		if (*searchlist!=';') isfragmentflag= 1L;
		}/* special-handling for MACROS */
else if (rdlevel>0) *brkstr= PNULL; 
else	{/* report error! */
	s= " ";
	if (nlevels>0) for (i=nlevels;i<=1;i--)
		s= catlist(8L,GC_PUSHS(s),GC_PUSHS("at nLevels="),cvs(i),
        GC_PUSHS(" searchList=\'"),GC_PUSHS(*aChkC(schlist,1L,i)),
        GC_PUSHS("\'"),GC_PUSHS(crlf),GC_PUSHS(" "));
	err_hlr(catlist(29L,GC_PUSHS("[COPY!EITHER] didn\'t find \'"),
        GC_PUSHS(searchlist),
        GC_PUSHS("\' at END-OF-STATEMENT for nLevel="),cvs(nlevels),
        GC_PUSHS(":"),GC_PUSHS(crlf),GC_PUSHS(s),
        GC_PUSHS("Start block-level="),cvs(startblocklevel),
        GC_PUSHS(crlf),GC_PUSHS(" Bottom block-level="),
        cvs(bottomlevel),GC_PUSHS(crlf),GC_PUSHS(" Block-level="),
        cvs(blevel),GC_PUSHS(crlf),GC_PUSHS(" IFC-level="),
        cvs(ifclevel),GC_PUSHS(crlf),GC_PUSHS(" LEX-RECUR-level="),
        cvs(rclevel),GC_PUSHS(crlf),GC_PUSHS(" str=\'"),
        GC_PUSHS(str),GC_PUSHS("\'"),GC_PUSHS(crlf),
        GC_PUSHS(" saiStatement=\'"),GC_PUSHS(saistatement),
        GC_PUSHS("\'")));
	}/* report error! */


/*  ------------------------------------------- */
/*  [4.0] . . IF /NOERROR THEN KEEP GOING . . .  */
/*  ------------------------------------------- */
GC_EPOPS(dbug(010L,catlist(17L,GC_PUSHS("&[COPY!EITHER].4 FAILED bLevel="),cvs(blevel),
        GC_PUSHS(" ifcLevel="),cvs(ifclevel),GC_PUSHS(crlf),
        GC_PUSHS(" searchList=\'"),GC_PUSHS(searchlist),GC_PUSHS(
        "\'"),GC_PUSHS(crlf),GC_PUSHS(" brkStr=\'"),
        GC_PUSHS(*brkstr),GC_PUSHS("\' str=\'"),GC_PUSHS(str),
        GC_PUSHS("\'"),GC_PUSHS(crlf),GC_PUSHS(" isFRAGMENTflag="),
        pbool(isfragmentflag))));
/* NO-OP */;

if (isfragmentflag) *brkstr= PNULL;
popnumber(&iscopy_toflag);


/* NO-OP */;

;
{aryfree((INTEGER *)brklist,(INTEGER *)schlist,
        (INTEGER *)0);  /* Free List of Dynamic Arrays */
 GC_SRETURN(str);}			/*  -- Return what we have in any case!  */
/*************************************************/
/*        U N U S E D    D E C L A R A T I O N S */
/* 's2'=<L0> string                              */
/*************************************************/

#undef pop_level
#undef do_push_level
#undef chk_fix_pdls
#undef plvl
#undef maxlexcodes
#undef bjustchk
#undef bcomma
#undef bsemicolon
#undef brsquiggle
#undef brparen
#undef brbrkt
#undef bendc
#undef bend
#undef bwhile
#undef bdo
#undef bfor
#undef bdelim
#undef maxlevels
}/* COPY!EITHER */
