/* FILE: cv_print_arglist.c
                            Peter Lemkin 

                   Image Processing Section, DCBD
                      National Cancer Institute
                        FCRF Bld 469, Rm 150B
                         Frederick, MD 21701

			    June 7, 1987

     COPYRIGHT NOTICE - Copyright (c) 1983-1988 by Peter F. Lemkin.
                        All rights reserved.
	Permission is granted to any individual or institution to
        copy, use or modify this program, except for explicitly
        commercial purposes, provided this copyright notice is
        included in its entirety in all derived works.  No warranty
        of the software or of the accuracy of any documentation
        surrounding it is expressed or implied, and the author does
        not acknowledge any liability resulting from program or
        documentation errors.

	      Revised January 3, 1988
              Derived from PSAIL.SAI V.43.50   7-19-84
 */


/* File: pscvt.c, being converted 12/25/87, 03:15:56PM*/
/* PSAIL compiler, version: June 9, 1987 V.57.10(OPTIMIZED) 	*/

#include "pshead.h"	/* header file to include other header files */


/*  ------------------------------------------------------ */
 /*INTERNAL*/ STRING cv_print_arglist()
/*  ------------------------------------------------------ */
{/* CV!PRINT!ARGLIST */ 
/*   --
.LABEL [<++PSAIL PROCEDURE: CV!PRINT!ARGLIST>]
.INDEX ++PSAIL PROCEDURE: CV!PRINT!ARGLIST
DESCRIPTION:
	Given PRINT FSM argument list (in saiStatement), analyze it
by recursive descent and compute a format. Map non-scalar types to
INTEGER with a /*WARNING*/ code. All LEAP variables x map to
CVIS(x,0) to print their print names if they exist else the LEAP
number converted to a string by the <leaprun.c> cvis() runtime.
The conversion codes are:
	DATA TYPE	CONVERISON CODE
	---------	---------------
	'c'			%c
	"string const"		%s
	STRING			%s
	float			%g  (shorter of %f and %e)
	INTEGER			%l
	LEAP-type		%s   (uses cvis(x,0))

For example:
		Integer pCnt, nbr
		Real v
		String CRLF
		PRINT("These",v,"entries have ",pCnt," of ",nbr,CRLF)
produce:
  ("%s%g%s%l%s%l%s","These",v,"entries have ",pCnt," of ",nbr,CRLF)

@V.52.18 NOTE: maybe redo the decode IF-THEN-ELSE-IF chain as a table.
@V.53.59 NOTE: All string args are pushed onto the Temp String Var
PDL. However, the format control string is not pushed and MUST BE A
STRING CONSTANT - not a string variable!

SIDE-EFFECTS:

PROBLEMS/KNOWN-BUGS:

.ENDLABEL [<++PSAIL PROCEDURE: CV!PRINT!ARGLIST>]
 */
STRING format, 
	id, 
	expr, 
	type, 
	sresult, 
	brkstr, 
	cvcode, 
	nowsexpr;
INTEGER ch, 
	brk, 
	pcode, 
	thistype;


GC_EPOPS(dbug(0200000L,catlist(3L,
			       GC_PUSHS("[CV!PRINT!ARGLIST] saiStatement=\'"),
			       GC_PUSHS(saistatement),GC_PUSHS("\'"))));

GC_MARKSSTKFRAME;   /* Mark G.C. STACK-FRAME */
GC_NOTESTRING(format); /* for G.C. */
GC_NOTESTRING(id); /* for G.C. */
GC_NOTESTRING(expr); /* for G.C. */
GC_NOTESTRING(type); /* for G.C. */
GC_NOTESTRING(sresult); /* for G.C. */
GC_NOTESTRING(brkstr); /* for G.C. */
GC_NOTESTRING(cvcode); /* for G.C. */
GC_NOTESTRING(nowsexpr); /* for G.C. */
pushnumber(lasttype);
format= quote;
brkstr= PNULL;
sresult= ",";
while ((*saistatement!='\0') &&(*brkstr!=')'))
	{/* munch arg list */
	expr= GC_EPOPS(copy_either(GC_PUSHS(",)"),&brkstr));
	while ((*expr!='\0') &&iswhitespace((INTEGER)*expr))
	  ps_itrash_= LOP(expr);
	{/* TRASH!RIGHT!WS */
	    while ((*expr!='\0') &&
		   (ps_itrash_= (*(expr+length(expr)-1))) &&
		   iswhitespace(ps_itrash_)) 
	      expr= subst(GC_PUSHS(expr),1L,length(expr)-1L);
			}/* TRASH!RIGHT!WS */	
	/*  -- JUSTIFY!TEXT will put it back
					where it is needed */
/*  # # @V.53.88	PUSH(brkStr) */
/*  # # @V.53.88	PUSH(sResult) */
/*  # # @V.53.88	PUSH(format) */
	midofexpr= 1L;
	lasttype= 0L;		/*  -- i.e. there is none!!! */
	pushnumber(usefoldingflag); usefoldingflag= 0L;
				/*  -- we need this to avoid trashing
				     things like CRLF! */
	nowsexpr= expr= GC_EPOPS(r_d_expr_map(GC_PUSHS(expr)));
	popnumber(&usefoldingflag);
/*  # # @V.53.88	POP(format) */
/*  # # @V.53.88	POP(sResult) */
/*  # # @V.53.88	POP(brkStr) */
	pcode= (((ps_itrash_= (lasttype&037))>0) ? (ps_itrash_+0140L)
				:0L);
/*  # # DBUG('200000,"[CV!PRINT!ARGLIST].1 lastType='"&CVTYPES(lastType)&
    "'"&CRLF&"  expr='"&expr&"' pCode='"&(ps!sTrash!_pCode)&"'") */

	while ((*nowsexpr!='\0') &&iswhitespace((INTEGER)*nowsexpr))
	  ps_itrash_= LOP(nowsexpr);	/*  -- for %c test */
	if ((*nowsexpr==*singlequote)) cvcode= "%c"; 
	else if (pcode==codestring) 	{/* add GC_PUSHS(expr) */
			cvcode= "%s";
			add_gc_pushs(&expr);
			}/* add GC_PUSHS(expr) */
	else if (pcode==codereal) cvcode= "%g"; 
	else if (pcode==codedouble) cvcode= "%d"; 
	else if (pcode==codeinteger) cvcode= "%l"; 
	else if (lasttype&020000L)
		{/* map to LEAP */
		cvcode= "%s";
		expr= catlist(3L,GC_PUSHS("cvis("),GC_PUSHS(expr),
			      GC_PUSHS(",0)"));
		pdw_ptype(lasttype,codestring);
		}/* map to LEAP */
	else cvcode= PNULL;
	if (*cvcode=='\0')
		{/* map to INTEGER */
		thistype= 0;
		pdw_ptype(thistype,codeinteger);
		expr= GC_EPOPS(cvactual_to_expected(thistype,lasttype,
						    GC_PUSHS(expr)));
		lasttype= thistype;
		cvcode= "%l";
		err_warning(&cvcode);
		}/* map to INTEGER */
	format= catlist(2L,GC_PUSHS(format),GC_PUSHS(cvcode));
	sresult= catlist(3L,GC_PUSHS(sresult),GC_PUSHS(expr),GC_PUSHS(","));
/*  # # DBUG('200000,"[CV!PRINT!ARGLIST].2 lastType='"&CVTYPES(lastType)&
    "'"&CRLF&"  format='"&format&"'"&CRLF&" sResult='"&sResult&"'") */
	}/* munch arg list */

format= catlist(2L,GC_PUSHS(format),GC_PUSHS(quote));
sresult= catlist(4L,GC_PUSHS("("),GC_PUSHS(format),
	    subst(GC_PUSHS(sresult),1L,length(sresult)-1L),GC_PUSHS(")"));
GC_EPOPS(dbug(0200000L,catlist(3L,
			 GC_PUSHS("[COMPUTE!PRINT!ARGLIST].3 sResult=\'"),
			       GC_PUSHS(sresult),GC_PUSHS("\'"))));
popnumber(&lasttype);

GC_SRETURN(sresult)
/*************************************************/
/*        U N U S E D    D E C L A R A T I O N S */
/* 'brk'=<L0> integer                            */
/* 'ch'=<L0> integer                             */
/* 'type'=<L0> string                            */
/* 'id'=<L0> string                              */
/*************************************************/

}/* CV!PRINT!ARGLIST */
