/* FILE: cvactual_to_expected.c
                            Peter Lemkin 

                   Image Processing Section, DCBD
                      National Cancer Institute
                        FCRF Bld 469, Rm 150B
                         Frederick, MD 21701

			    June 7, 1987

     COPYRIGHT NOTICE - Copyright (c) 1983-1988 by Peter F. Lemkin.
                        All rights reserved.
	Permission is granted to any individual or institution to
        copy, use or modify this program, except for explicitly
        commercial purposes, provided this copyright notice is
        included in its entirety in all derived works.  No warranty
        of the software or of the accuracy of any documentation
        surrounding it is expressed or implied, and the author does
        not acknowledge any liability resulting from program or
        documentation errors.

	      Revised January 3, 1988
              Derived from PSAIL.SAI V.43.50   7-19-84
 */


/* File: pscvt.c, being converted 12/25/87, 03:15:56PM*/
/* PSAIL compiler, version: June 9, 1987 V.57.10(OPTIMIZED) 	*/

#include "pshead.h"	/* header file to include other header files */


/*  ------------------------------------------------------ */
/*INTERNAL*/  STRING cvactual_to_expected(dwexpected,dwactual,actualexpr)
	INTEGER dwexpected;
	INTEGER dwactual;
	STRING actualexpr;
/*  ------------------------------------------------------ */
{/* CVACTUAL!TO!EXPECTED */
/*   --
.LABEL [<++PSAIL PROCEDURE: CVACTUAL!TO!EXPECTED>]
.INDEX ++PSAIL PROCEDURE: CVACTUAL!TO!EXPECTED
DESCRIPTION:

SIDE-EFFECTS:

PROBLEMS/KNOWN-BUGS:

.ENDLABEL [<++PSAIL PROCEDURE: CVACTUAL!TO!EXPECTED>]
	NOTE: @V.52.122 - If we are trying to coerce a Macro formal
argument which did NOT have a type parameter (i.e. DW!macro and
DW!formalArg) when isMACROflag is true (i.e. doing R!D! on macro
body) then set isFRAGMENTflag.

	If the expected and actual argument types are the same the
simply return the actual argument. If mapping numeric constants then
do integer<==>real rather than use casts. Otherwise, transform it
according to the following table where:

.LABEL [<PSAIL-COERCION-TBL>]
	Let e be the actualExpr,
	Let R be the REAL mapping (float) <==> (double) cast if any,
	Let M be the makstr(...),
	Let L be the CVI(...),
	Let N be the CVN(...),
	Let b be the (BYTE) cast - i.e. (unsigned char),
	Let i be the (INTEGER) cast,
	Let f be the (float) or (double) cast,
	Let c be the character '*' access,
	Let w be /*WARNING*/,
	Let X be the ERROR.

                                  i
                 d   s      r     t                         o
                 o   t      C     e                   e     b
               r u   r    r l   i m          l    p   v     j   l    b
       E  i    e b   i    P a   t v    s     i    r   e     e   i    y
       X  n    a l   n    t s   e a    e     s    o   n     c   s    t
   ACT\P  t |  l e|  g  | r s | m r |  t  |  t  | c | t  |  t | p  | e
----------------------------------------------------------------------
1  int |  e |  fe |  Me |  ew | Le  |  Le |  Le |ew | ew | ew | ew | be
2  real| ie |  R  | Mie |  iew| Lie | Lie |  Lie|Xe |iew | ew | ew | be
3  str | ce | fce |   e | Xcew| XLce| XLce| Lce |Xe |Xcew|Xcew|Xcew| bce
4  rPtr| ew | few | XMew|   e |  ew |  ew |  ew |Xe | ew | ew | ew |Xbew
5  ivar| Ne | fNe | XMNe|  ew |   e |  ew |  ew |Xe | ew | ew | ew |Xbew
6  set | Ne | fNe | XMNe|  ew |  ew |   e |  ew |Xe | ew | ew | ew |Xbew
7  list| Ne | fNe | XMNe|  ew |  ew |  ew |  e  |Xe | ew | ew | ew |Xbew
8  proc|  e |  fe | XMe | Xew | Xew | Xew | Xew |e  | ew | ew | ew |Xbew
9  evnt| ew | few | XMew|  ew |  ew |  ew | ew  |ew | e  | ew | ew |Xbew
10 objt| ew | few | XMew|  ew |  ew |  ew | ew  |ew | ew | e  | ew |Xbew
11 lisp| ew | few | XMew|  ew |  ew |  ew | ew  |ew | ew | ew | e  |Xbew
12 byte| ie |  fe |  Me |  iew|  iew|  iew| iew |iew| iew| iew| ie |  e

The table is used to drive a CASE statement with cases for [e], [fe],
[Me], ... etc. which actually perform the required conversion.
.ENDLABEL [<PSAIL-COERCION-TBL>]

SIDE-EFFECTS:

PROBLEMS/KNOWN-BUGS:

.ENDLABEL [<++PSAIL PROCEDURE: CVACTUAL!TO!EXPECTED>]
 */

/*  -- List of operation codes --  */
#define x 0
#define e 1
#define r 2
#define fe 3
#define me 4
#define ew 5
#define le 6
#define ie 7
#define mie 8
#define iew 9
#define lie 10
#define xe 11
#define ce 12
#define fce 13
#define xcew 14
#define xlce 15
#define xmew 16
#define xew 17
#define xme 18
#define few 19
#define lce 20
#define ne 21
#define fne 22
#define xmne 23
#define be 24
#define bce 25
#define xbew 26

#define maxtype 12

 static INTEGER /*ARRAY*/ P_mapae[16+(12-1+1L)*(12-1+1L)] = { 
   0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, e,  fe,   me,   ew,  le,    le,
      le, ew,  ew,  ew,  ew, be, /*  1Int */
   ie,   r,  mie,   iew, lie,  lie,  lie, xe, iew,  ew,  ew,  be,
    /*  2Real */
   ce, fce,    e,  xcew, xlce, xlce, lce, xe, xcew,xcew,xcew, bce,
    /*  3Str */
   ew, few,  xmew,   e,   ew,   ew,   ew, xe,  ew,  ew,  ew, xbew,
    /*  4Rec */
   ne, fne,  xmne,  ew,    e,   ew,   ew, xe,  ew,  ew,  ew, xbew,
   /*  5Item */
   ne, fne,  xmne,  ew,   ew,    e,  ew,  xe,  ew,  ew,  ew, xbew,
    /*  6Set */
   ne, fne,  xmne,  ew,   ew,   ew,   e,  xe,  ew,  ew,  ew, xbew,
   /*  7List */
    e,  fe,  xme,  xew,  xew,  xew,  xew, e,   ew,  ew,  ew, xbew,
   /*  8Proc */
   ew, few,  xmew,  ew,   ew,   ew,  ew,  ew,  e,   ew,  ew, xbew,
   /*  9Evnt */
   ew, few,  xmew,  ew,   ew,   ew,  ew,  ew,  ew,  e,   ew,xbew,
    /*  10Obj */
   ew, few,  xmew,  ew,   ew,   ew,  ew,  ew,  ew,  ew,  e, xbew,
   /*  11Lisp */
   ie, fe,   me,   iew,  iew,  iew, iew, iew, iew, iew, ie,  e };
 static INTEGER /*ARRAY*/ *mapae  /* [1:maxtype,1:maxtype] */ ;
 /*  -- same tbl as above */

 static INTEGER /*ARRAY*/ P_pmapi[13+(31-0+1L)] = { 
   0,0,0,0,0,0,0,0,0,0,0,0,0, 0,	/*  -- NO-CODE = 0 'NULL' */
   	0,	/*  -- ARRAY = 1** array 'a' */
   	0,	/*  -- INTERRUPT = 2 interrupt 'b' */
   	3,	/*  -- STRING = 3 string 'c' */
   	2,	/*  -- LONG REAL = 4 double 'd' */
   	9,	/*  -- EVENT = 5 event 'e' */
   	2,	/*  -- REAL = 6 float 'f' */
   	0,	/*  -- LABEL = 7 label 'g' */
   	11,	/*  -- NONE = 8 lisp!atom 'h' */
   	1,	/*  -- INTEGER or BOOLEAN = 9 integer 'i' */
   	10,	/*  -- NONE = 10 object 'j' */
   	11,	/*  -- NONE = 11 lisp!S!expr 'k' */
   	7,	/*  -- LIST = 12 list 'l' */
   	0,	/*  -- NONE = 13** macro 'm' */
   	8,	/*  -- PROCEDURE = 14 void 'n' */
   	0,	/*  -- NONE = 15** optional 'o' */
   	8,	/*  -- PROCEDURE = 16 process 'p' */
   	4,	/*  -- RECORD!POINTER = 17 record!pointer 'q' */
   	0,	/*  -- REFERENCE =  18** reference 'r' */
   	6,	/*  -- SET = 19 set 's' */
   	4,	/*  -- RECORD!CLASS = 20 record!class 't' */
   	0,	/*  -- NONE = 21 unspecified 'u' */
   	5,	/*  -- ITEMVAR = 22 itemvar 'v' */
   	0,	/*  -- NONE = 23** own(static) 'w' */
   	5,	/*  -- ITEM = 24 item 'x' */
   	0,	/*  -- NONE = 25 record!field 'y' */
   	0,	/*  -- NONE = 26** safe 'z' */
   	0,	/*  -- NONE = 27** Item!array '{' */
   	1,	/*  -- NONE = 28** short '|' -- same as INTEGER */
   	12,	/*  -- NONE = 29** byte '}' -- i.e. (unsigned char) */
   	0,	/*  -- NONE = 30** C!array '~' */
   	1 };
 static INTEGER /*ARRAY*/ *pmapi  /* [0:31] */ ;
 /*  -- map the pCode Land '37 against
				possible entries in this table which
				are legitimate data types since the
				primary codes can have other stuff.
				The value generated is in index
				jA or jE for the mapAE[jA,jE] table.
				It is used by 'pMapI[dw Land '37]' etc.
				 */

STRING s, 
	s1, 
	s2, 
	mexpr;
INTEGER ch, 
	i, 
	j, 
	cactual, 
	cexpected, 
	val, 
	ja, 
	je, 
	didmakstr;


GC_EPOPS(dbug(02L,catlist(11L,GC_PUSHS("[CVACTUAL!TO!EXPECTED] actualExpr=\'"),
        GC_PUSHS(actualexpr),GC_PUSHS("\'"),GC_PUSHS(crlf),
        GC_PUSHS("  dwActual=\'"),cvtypes(dwactual),GC_PUSHS("\'"),
        GC_PUSHS(crlf),GC_PUSHS("  dwExpected=\'"),cvtypes(
        dwexpected),GC_PUSHS("\'"))));


/*  ---------------------------------------------- */
/*  . . . PRE-PROCESS DW's PRIOR TO MAPPING . . .  */
/* IFC EVAL(not ps_optimize)=1 */
/* THENC */

if ((dwactual&0100L) ||(dwexpected&0100L)) err_hlr(GC_PUSHS(

         	catlist(GC_PUSHS(
        "[CVACTUAL!TO!EXPECTED] DRYROT - used CODExxx instead of PDW!pType()")
        ,GC_PUSHS(crlf),GC_PUSHS("actualExpr=\'"),GC_PUSHS(
        actualexpr),GC_PUSHS("\'"),GC_PUSHS(crlf),
        GC_PUSHS("dwActual=\'"),cvtypes(dwactual),GC_PUSHS("\'"),
        GC_PUSHS(crlf),GC_PUSHS("dwExpected=\'"),cvtypes(dwexpected)
        ,GC_PUSHS("\'"))));
 /* ENDC */

/*  ---------------------------------------------- */
GC_POPS(actualexpr); /* from GC string PDL */
GC_MARKSSTKFRAME;   /* Mark G.C. STACK-FRAME */
GC_NOTESTRING(actualexpr); /* for G.C. */
GC_NOTESTRING(s); /* for G.C. */
GC_NOTESTRING(s1); /* for G.C. */
GC_NOTESTRING(s2); /* for G.C. */
GC_NOTESTRING(mexpr); /* for G.C. */
mapae = aryHdrI(P_mapae,2,1,12,1,12,"mapae",'i');	/* Dynamic Array */
pmapi = aryHdrI(P_pmapi,1,0,31,"pmapi",'i');	/* Dynamic Array */
if (!usecoerceflag ||((((ps_itrash_= (dwexpected&037))>0) ? (ps_itrash_+0140L)
				:0L)==codeuntyped)) 
	{/* do NOTHING if /NOCOERCE!!! */
/*  # # DBUG('2,"[CVACTUAL!TO!EXPECTED].0.0.1 useCOERCEflag="&PBOOL(useCOERCEflag)& " dwExpected Land '37=0"&CVOS(dwExpected Land '37)) */
	;
	{GC_UNMARKSSTKFRAME;
	 GC_SRETURN(actualexpr);}
	}/* do NOTHING if /NOCOERCE!!! */

if ((((ps_itrash_= (dwactual&037))>0) ? (ps_itrash_+0140L)
				:0L)==0&&!ismacroflag &&
   !equ(subst(GC_PUSHS(actualexpr),length(actualexpr)-10L,length(actualexpr)),GC_PUSHS("/*WARNING*/"))) 
	{/* make a /*WARNING*/!!! */
/*  # # DBUG('2,"[CVACTUAL!TO!EXPECTED].0.0.2 isMACROflag="&PBOOL(isMACROflag)& " dwActual Land '37=0"&CVOS(dwActual Land '37)&CRLF&" actualExpr='"&actualExpr&"'") */
	err_warning(&actualexpr);
	;
	{GC_UNMARKSSTKFRAME;
	 GC_SRETURN(actualexpr);}
	}/* make a /*WARNING*/!!! */


/*  --------------------------------------------------------- */
/*  . . . HANDLE MACRO FRAGMENTS WHICH HAVE 'UNTYPED' IDS . . */
/*  --------------------------------------------------------- */
/*  -- Code to detect MACRO FRAGMENTS with 'untyped' formal macro params
during R!D of the macro body while defining the macro in FSM1.
 */
	if (ismacroflag &&(dwactual&040000L) &&
	   (dwactual&02000L))
		{/* Do NOT coerce untyped macro formal args */
GC_EPOPS(dbug(02L,catlist(2L,GC_PUSHS("[CVACTUAL!TO!EXPECTED].0.0.3 isFRAGMENTflag="),
        pbool(isfragmentflag))));
		isfragmentflag= 1L;   /*  -- ==> PSAIL!EVAL */
		{GC_UNMARKSSTKFRAME;
		 GC_SRETURN(actualexpr);}
		}/* Do NOT coerce untyped macro formal args */

push(GC_PUSHS(ps_warning_));		/*  -- NOTE: no warnings in MACROS... */
if (ismacroflag) ps_warning_= PNULL;


/*  ----------------------------------------- */
/* 	. . . DO PRAGMA SPECIAL CASES . . . */
/*  ----------------------------------------- */
if (*actualexpr=='('&&(*(actualexpr+length(actualexpr)-1))==')')
	{/* try to simplify (..) levels */
	s1= subst(GC_PUSHS(actualexpr),2L,length(actualexpr)-1L);
	if (chk_id(GC_PUSHS(s1))) 
		{/* Yes */
		actualexpr= s1;	/*  -- trash extra (...)  */
/*  # # DBUG('2,"[CVACTUAL!TO!EXPECTED].0.1 actualExpr='"&actualExpr&"'") */
		}/* Yes */
	}/* try to simplify (..) levels */


/*  ----------------------------------------- */
/*  . . . DO /MATH MAPPING IF NECESSARY . . . */
/*  ----------------------------------------- */
if (((((ps_itrash_= (dwexpected&037))>0) ? (ps_itrash_+0140L)
				:0L)==codedouble&&usepmathflag) ||
   (*promotedouble=='\0')) pdw_ptype(dwexpected,codereal);
if (((((ps_itrash_= (dwactual&037))>0) ? (ps_itrash_+0140L)
				:0L)==codedouble&&usepmathflag) ||
   (*promotedouble=='\0')) pdw_ptype(dwactual,codereal);

GC_EPOPS(dbug(02L,catlist(11L,GC_PUSHS("[CVACTUAL!TO!EXPECTED].0.2 actualExpr=\'"),
        GC_PUSHS(actualexpr),GC_PUSHS("\'"),GC_PUSHS(crlf),
        GC_PUSHS("  dwActual=\'"),cvtypes(dwactual),GC_PUSHS("\'"),
        GC_PUSHS(crlf),GC_PUSHS("  dwExpected=\'"),cvtypes(
        dwexpected),GC_PUSHS("\'"))));


/*  ------------------------------------------------------------------ */
/*  . . . EXTRACT PRIMARY TYPE CODES AND DO THE REQUIRED MAPPING . . . */
/*  ------------------------------------------------------------------ */
cactual= (((ps_itrash_= (dwactual&037))>0) ? (ps_itrash_+0140L)
				:0L);    /*  -- get primary data type */
cexpected= (((ps_itrash_= (dwexpected&037))>0) ? (ps_itrash_+0140L)
				:0L);
ja= (*aChkI(pmapi,1L,dwactual&037));	/*  -- get local type table index */
je= (*aChkI(pmapi,1L,dwexpected&037));
/*  # # DBUG('2,"[CVACTUAL!TO!EXPECTED].1 jA="&CVS(jA)&" jE="&CVS(jE)&CRLF&" cActual='"&(ps!sTrash!_cActual)&"'"&CRLF&"  cExpected='"&(ps!sTrash!_cExpected)&"'") */


/*  ------------------------------------------------------------------ */
/*  . . . FORCE 32-bit INTEGERS IF ACTUAL EXPR IS INTEGER NUMBER . . . */
/*  ------------------------------------------------------------------ */
/*  ------------------------------------------------------------------ */
if ((ja==1L&&je==1L) /*  # # @V.54.04 And ISDIGIT(actualExpr)  */ &&
	chk_integer(GC_PUSHS(actualexpr)))
	{/* force 32-bit integer */
	actualexpr= ((bitspsail&0400000L) ? 
			catlist(2L,GC_PUSHS("(INTEGER)"),GC_PUSHS(actualexpr)): catlist(2L,GC_PUSHS(actualexpr),GC_PUSHS("L")));
GC_EPOPS(dbug(02L,catlist(5L,GC_PUSHS("[CVACTUAL!TO!EXPECTED].1.0.1 bitsPSAIL="),
        cvos(bitspsail),GC_PUSHS(" actualExpr=\'"),
        GC_PUSHS(actualexpr),GC_PUSHS("\'"))));
	}/* force 32-bit integer */


/*  ------------------------------------------------------------- */
/*  . . . MAP ACTUAL TO EXPECTED DOING WHATEVER IS REQUIRED . . . */
/*  ------------------------------------------------------------- */
didmakstr= 0L;		/*  -- assume not [M] conversion */
/*  -- @V.52.194 Changed to new table method -- */
i= (((ja==0||je==0)) ?   0L: (*aChkI(mapae,2L,ja,je)));
/*  # # DBUG('2,"[CVACTUAL!TO!EXPECTED].1.1 i="&CVS(i)&" jA="&CVS(jA)&" jE="&CVS(jE)) */
switch (i)
	{/* map actualExpr to mExpr */
case 0L: 	{/* Fatal Error */
	if ((i==0) &&ismacroflag) isfragmentflag= 1L;  /*  -- ABORT THIS MESS -- */
	mexpr= catlist(2L,GC_PUSHS("@ERR@"),GC_PUSHS(actualexpr));; break;
	}/* Fatal Error */

case 1L: 	{/* 1:1 mapping */
	mexpr= ((*actualexpr==*singlequote) &&(cactual==codeinteger)) ? 
			catlist(2L,GC_PUSHS("(INTEGER)"),GC_PUSHS(actualexpr)): actualexpr;; break;
	}/* 1:1 mapping */

case 2L: 	{/* (float) or (double) if required */
	mexpr= ((cactual==cexpected) ? actualexpr
		: catlist(2L,GC_PUSHS((cexpected==codedouble) ? "(double)"
        			: "(float)"),GC_PUSHS(actualexpr)));; break;
	}/* (float) or (double) if required */

case 3L: 	{/* (REAL) cast */
	mexpr= catlist(2L,GC_PUSHS((cexpected==codedouble) ? "(double)" : "(float)"),
        GC_PUSHS(actualexpr));; break;
	}/* (REAL) cast */

case 4L: 	{/* makstr(expr) */
	/*  -- test for special 'c' case */
	if (chk_integer(GC_PUSHS(actualexpr)) &&
		((*actualexpr=='0'&&
		  (00<=(val= cvo(GC_PUSHS(actualexpr))) && val<=0177)) ||
		 (00<=(val= cvd(GC_PUSHS(actualexpr))) && val<=0177)))	{/* make 3-digit number '\000' */
			INTEGER p,
			q;
			getformat(&p,&q);
			setformat(-3L,0L);  /*  -- 3 digits leading zeros */
			mexpr= catlist(4L,GC_PUSHS(quote),GC_PUSHS("\\"),cvos(val),GC_PUSHS(quote));
			setformat(p,q);
			}/* make 3-digit number '\000' */
	else if (*actualexpr==('\''/*WARNING*/)&&(*(actualexpr+length(actualexpr)-1))==('\''/*WARNING*/)) mexpr= catlist(3L,GC_PUSHS(quote),subst(GC_PUSHS(actualexpr),2L,
        length(actualexpr)-1L),GC_PUSHS(quote)); else 	{/* string fct */
		mexpr= catlist(3L,GC_PUSHS("makstr("),GC_PUSHS(actualexpr),GC_PUSHS(")"));
		didmakstr= 1L;
		}/* string fct */; break;
	}/* makstr(expr) */

case 5L: 	{/* expr/&WARNING */
	mexpr= actualexpr;
	err_warning(&mexpr);; break;
	}/* expr/&WARNING */

case 6L: 	{/* LEAP */
	mexpr= catlist(3L,GC_PUSHS("cvi("),GC_PUSHS(actualexpr),GC_PUSHS(")"));; break;
	}/* LEAP */

case 7L: 	{/* (INTEGER) cast */
	mexpr= catlist(2L,GC_PUSHS("(INTEGER)"),GC_PUSHS(actualexpr));; break;
	}/* (INTEGER) cast */

case 8L: 	{/* makstr((INTEGER)<expr>) */
	mexpr= catlist(3L,GC_PUSHS("makstr((INTEGER)"),GC_PUSHS(actualexpr),
        GC_PUSHS(")"));; break;
	}/* makstr((INTEGER)<expr>) */

case 9L: 	{/* (INTEGER)<expr>&WARNG */
	mexpr= catlist(2L,GC_PUSHS("(INTEGER)"),GC_PUSHS(actualexpr));
	err_warning(&mexpr);; break;
	}/* (INTEGER)<expr>&WARNG */

case 10L: 	{/* cvi((INTEGER)<expr>) */
	mexpr= catlist(3L,GC_PUSHS("cvi((INTEGER)"),GC_PUSHS(actualexpr),GC_PUSHS(")"));; break;
	}/* cvi((INTEGER)<expr>) */

case 11L: 	{/* @ERR@&<expr> */
	mexpr= catlist(2L,GC_PUSHS("@ERR@"),GC_PUSHS(actualexpr));; break;
	}/* @ERR@&<expr> */

case 25L: 
case 12L: 	{/* [ce]:STRING==>INTEGER, [bce]:STRING==>BYTE */
	/*  -- test for special 'c' case */
	/*  -- For INTEGER we preface a (INTEGER) case */
	if (fast_equ(GC_PUSHS(actualexpr),GC_PUSHS("PNULL"))) mexpr= "\'\\0\'"; 
	else if (*actualexpr==('\''/*WARNING*/)&&(*(actualexpr+length(actualexpr)-1))==('\''/*WARNING*/)) mexpr= actualexpr; 
	else if (*actualexpr==*quote&&(*(actualexpr+length(actualexpr)-1))==*quote) 
		{/* optmize to 'c' */
		/*  -- if have string "ABC...Z" then
		  -- map to 
			('A'WARNING)
		 */
		s2= actualexpr;
		ch= LOP(s2);	/*  -- trash first quote */
		s1= GC_EPOPS(makstr(LOP(s2)));	/*  -- get char */
		if (*s1==('\\'/*WARNING*/)) 
			{/* get rest of it */
			if (!isoctal((INTEGER)*s2)) s1= catlist(2L,GC_PUSHS(s1),GC_PUSHS(LOP(s2))); else	do {s1= catlist(2L,GC_PUSHS(s1),GC_PUSHS(LOP(s2)));} while (!(!isoctal((INTEGER)*s2)));
			}/* get rest of it */
		mexpr= catlist(3L,GC_PUSHS("(\'"),GC_PUSHS(s1),GC_PUSHS("\'"));
		err_warning(&mexpr);
		mexpr= catlist(2L,GC_PUSHS(mexpr),GC_PUSHS(")"));
		if (*s1=='\0') mexpr= 
        					catlist(GC_PUSHS("(*"),GC_PUSHS(actualexpr),GC_PUSHS(
        ")")); /*  -- OPS! */
		}/* optmize to 'c' */
	else 	{/* put in *(...) */
		if (chk_id(GC_PUSHS(actualexpr))) mexpr= catlist(2L,GC_PUSHS("*"),GC_PUSHS(actualexpr)); else mexpr= catlist(3L,GC_PUSHS("*("),GC_PUSHS(actualexpr),GC_PUSHS(")"));
			}/* put in *(...) */
	if (i==12L) mexpr= catlist(2L,GC_PUSHS("(INTEGER)"),GC_PUSHS(mexpr));; break;
	}/* [ce]:STRING==>INTEGER, [bce]:STRING==>BYTE */

case 13L: 	{/* (REAL)*<expr> */
	mexpr= 
        		catlist(
        GC_PUSHS((cactual==codedouble) ? "(double)" : "(float)(*"),
        GC_PUSHS(actualexpr),GC_PUSHS(")"));; break;
	}/* (REAL)*<expr> */

case 14L: 	{/* @ERR@ *<expr> & WARNING */
	mexpr= catlist(2L,GC_PUSHS("@ERR@*"),GC_PUSHS(actualexpr));
	err_warning(&mexpr);; break;
	}/* @ERR@ *<expr> & WARNING */

case 15L: 	{/* @ERR@ cvi(*<expr>) */
	mexpr= catlist(3L,GC_PUSHS("@ERR@cvi(*"),GC_PUSHS(actualexpr),GC_PUSHS(")"));; break;
	}/* @ERR@ cvi(*<expr>) */

case 16L: 	{/* @ERR@ makstr(<expr>) WARNING */
	mexpr= catlist(3L,GC_PUSHS("@ERR@makstr("),GC_PUSHS(actualexpr),GC_PUSHS(")"));
	err_warning(&mexpr);
	didmakstr= 1L;; break;
	}/* @ERR@ makstr(<expr>) WARNING */

case 17L: 	{/* @ERR@ <expr> WARNING */
	mexpr= catlist(2L,GC_PUSHS("@ERR@"),GC_PUSHS(actualexpr));
	err_warning(&mexpr);; break;
	}/* @ERR@ <expr> WARNING */

case 18L: 	{/* @ERR@ makstr(<expr>) */
	mexpr= catlist(3L,GC_PUSHS("@ERR@makstr("),GC_PUSHS(actualexpr),GC_PUSHS(")"));
	didmakstr= 1L;; break;
	}/* @ERR@ makstr(<expr>) */

case 19L: 	{/* (REAL) <expr> WARNING */
	mexpr= catlist(2L,GC_PUSHS((cexpected==codedouble) ? "(double)" : "(float)"),
        GC_PUSHS(actualexpr));
	err_warning(&mexpr);; break;
	}/* (REAL) <expr> WARNING */

case 20L: 	{/* cvi(*<expr>) */
	mexpr= catlist(3L,GC_PUSHS("cvi(*"),GC_PUSHS(actualexpr),GC_PUSHS(")"));; break;
	}/* cvi(*<expr>) */

case 21L: 	{/* cvn(<expr>) */
	mexpr= catlist(3L,GC_PUSHS("cvn("),GC_PUSHS(actualexpr),GC_PUSHS(")"));; break;
	}/* cvn(<expr>) */

case 22L: 	{/* (REAL)cvn(<expr>) */
	mexpr= 
        				catlist(
        GC_PUSHS((codedouble==cactual) ? "(double)" : "(float)cvn(")
        ,GC_PUSHS(actualexpr),GC_PUSHS(")"));; break;
	}/* (REAL)cvn(<expr>) */

case 23L: 	{/* @ERR@ makstr(cvn(<expr>)) */
	mexpr= catlist(3L,GC_PUSHS("@ERR@makstr(cvn("),GC_PUSHS(actualexpr),
        GC_PUSHS("))"));
	didmakstr= 1L;; break;
	}/* @ERR@ makstr(cvn(<expr>)) */

case 24L: 	{/* Cast to BYTE */
	mexpr= ((*actualexpr==*singlequote) &&(cactual==codeinteger)) ? 
			actualexpr
	       : catlist(3L,GC_PUSHS("(unsigned char)("),GC_PUSHS(actualexpr),
        GC_PUSHS(")"));
	/*  -- Note: we could eliminate the inner () if can figure out
		that it is a single identifier */; break;
	}/* Cast to BYTE */

case 26L: 	{/* @ERR@ *(<expr>) */
	mexpr= catlist(3L,GC_PUSHS("@ERR@*("),GC_PUSHS(actualexpr),GC_PUSHS(")"));; break;
	}/* @ERR@ *(<expr>) */
	}/* map actualExpr to mExpr */

/*  # # DBUG('2,"[CVACTUAL!TO!EXPECTED].2 mExpr='"&mExpr&"'") */

if (*mexpr=='@')
	{/* analyze @ERR@ error */
	mexpr= subst(GC_PUSHS(mexpr),6L,length(mexpr));
	if (!usechkflag) err_n_p(&mexpr); 
	else 	err_hlr(GC_PUSHS(
		
		        			catlist(GC_PUSHS(
		        "[CVACTUAL!TO!EXPECTED] Illeg. coercion of ["),
		        GC_PUSHS(actualexpr),GC_PUSHS("]"),GC_PUSHS(crlf),
		        GC_PUSHS(" Actual type=\'"),cvtypes(dwactual),GC_PUSHS(
		        "\'"),GC_PUSHS(crlf),GC_PUSHS(" Expected type=\'"),
		        cvtypes(dwexpected),GC_PUSHS("\'"),GC_PUSHS(crlf),
		        GC_PUSHS(" Final Expr=\'"),subst(GC_PUSHS(mexpr),6L,
		        length(mexpr)),GC_PUSHS("\'"))));
	}/* analyze @ERR@ error */

if (didmakstr)
	{/* did makstr() - so post process it if needed */
	sawstrfct= 1L;
	if (usegcflag) mexpr= catlist(3L,GC_PUSHS("GC_EPOPS("),GC_PUSHS(mexpr),GC_PUSHS(")"));
	}/* did makstr() - so post process it if needed */
GC_EPOPS(dbug(02L,catlist(6L,GC_PUSHS("[CVACTUAL!TO!EXPECTED].3 mExpr=\'"),GC_PUSHS(
        mexpr),GC_PUSHS("\' jA="),cvs(ja),GC_PUSHS(" jE="),cvs(je))));
pop(&ps_warning_);


GC_SRETURN(mexpr)
/*************************************************/
/*        U N U S E D    D E C L A R A T I O N S */
/* 'j'=<L0> integer                              */
/* 's'=<L0> string                               */
/*************************************************/


#undef maxtype
#undef xbew
#undef bce
#undef be
#undef xmne
#undef fne
#undef ne
#undef lce
#undef few
#undef xme
#undef xew
#undef xmew
#undef xlce
#undef xcew
#undef fce
#undef ce
#undef xe
#undef lie
#undef iew
#undef mie
#undef ie
#undef le
#undef ew
#undef me
#undef fe
#undef r
#undef e
#undef x
}/* CVACTUAL!TO!EXPECTED */
