/* FILE: fold_const_expr.c
                            Peter Lemkin 

                   Image Processing Section, DCBD
                      National Cancer Institute
                        FCRF Bld 469, Rm 150B
                         Frederick, MD 21701

			    June 7, 1987

     COPYRIGHT NOTICE - Copyright (c) 1983-1988 by Peter F. Lemkin.
                        All rights reserved.
	Permission is granted to any individual or institution to
        copy, use or modify this program, except for explicitly
        commercial purposes, provided this copyright notice is
        included in its entirety in all derived works.  No warranty
        of the software or of the accuracy of any documentation
        surrounding it is expressed or implied, and the author does
        not acknowledge any liability resulting from program or
        documentation errors.

	      Revised January 3, 1988
              Derived from PSAIL.SAI V.43.50   7-19-84
 */


/* File: pscvt.c, being converted 12/25/87, 03:15:56PM*/
/* PSAIL compiler, version: June 9, 1987 V.57.10(OPTIMIZED) 	*/

#include "pshead.h"	/* header file to include other header files */


/*   --------------------------------------- */
/*INTERNAL*/ STRING fold_const_expr(si)
	STRING si;
/*   --------------------------------------- */
{/* FOLD!CONST!EXPR */
/*   --
.LABEL [<++PSAIL PROCEDURE: FOLD!CONST!EXPR>]
.INDEX ++PSAIL PROCEDURE: FOLD!CONST!EXPR
DESCRIPTION:
	Eval C arithmetic operator strings of the form 1*(10-2)*(5)
to a string which contains that value if there are ONLY decimal
numbers involved. Operands must be either a) integer or real
constants, b) quoted string constants in which case the first
character is treated as an integer or c) a macro symbol (without
arguments) which is evaluated and must be a) or b).

The algorithm inputs the string expression and uses a Polish-Stack
evaluation procedure with operator precedence coded in prec[]. The
result is turned back into a string. If the expression contains
symbols which are not Macro-constants, then return the original
string. The Algorithm assumes a WFF. 

Note: MACRO constants are evaulated by using the macro body stored in
the declSize[idx] string filed of the macro symbol name. If the
DW!fragment bit is set, then this must be evaled as well. If the
declNargs[idx]>0 then we can not really eval it so abort!

FINAL TYPE:
----------
If the number is successfully evaluated as a floating or integer
number, then that type is set into lastType and the DW!macro and
DW!fragment bits cleared.

DO NOT FOLD:
------------
	a) Reals -- at this time...
	b) Simple one-operand expressions which includes:
		b.1) Octal numbers (keep as octal numbers!).
		b.2) TRUE or FALSE macros.
		b.3) 'C' single character.
		b.4) "text string" string where we would normally use
			the first character.
		b.5) If we see a <comma> or <semicolon> operator.

@V.54.01 NOTE: Need to handle numbers of the form
	'(INTEGER)12345' 
and
	'12345L'.

SIDE-EFFECTS:

PROBLEMS/KNOWN-BUGS:

.ENDLABEL [<++PSAIL PROCEDURE: FOLD!CONST!EXPR>]
 */
/* LABEL  abortfolding */
INTEGER h, 
	opr, 
	opr1, 
	i, 
	i1, 
	i2, 
	p, 
	q, 
	valptr, 
	oprptr, 
	idx, 
	brk,
	brk2, 
	dw, 
	ct, 
	ht, 
	isoprflag, 
	ch, 
	maxvals, 
	ispclc;
INTEGER isconstexprflag, 
	sawoctalflag, 
	sawrealflag, 
	didmacroexpandflag;
float v, 
	v1, 
	v2, 
	vnum;
STRING s, 
	s1, 
	s2, 
	s3, 
	so, 
	a, 
	siold, 
	token, 
	trash, 
	lasttoken,
	nomoretext, 
	lctoken, 
	wsright;
#define maxoprpdl 15	/*  -- we give up trying if > 15 operators! */
#define maxvalpdl 30L
 INTEGER /*ARRAY*/ *oprpdl  /* [0:maxoprpdl] */ ; /*  -- operator PDL */
 float /*ARRAY*/ *valpdl  /* [0:maxvalpdl] */ ;	 /*  -- operand PDL */

/*  @V.51.74 -- The following checks are overkill in terms of checking
but leave it in while debugging PSAIL. */

/* IFC EVAL(is_sail)=0 */
 /* THENC <<<FAILED!!!>>>

define opush(x) = \Begin If (oprPtr_oprPtr+1)>maxOPRPDL
			Then
			Begin "error"
			isConstExprFlag_FALSE;
			Goto abortFOLDING;
# # # @V.52.49		ERR!HLR(
   "[FOLD!CONST!EXPR] OPUSH oprPDL ovf! Input='"&CRLF&
   "'" & sIold &"'");
			End "error"
		   Else oprPDL[oprPtr]_(x);
		   End\;
  */ /* ENDC */

/* IFC EVAL(is_psail)=1 */
/* THENC */

#define opush(x) {if ((++oprptr)>15L)\
			{"error"\
			isconstexprflag= 0L;\
			goto abortfolding;\
/*  _2 _2 Ev.52.49		err_hlr(\
   "[FOLD!CONST!EXPR] OPUSH oprPDL overflow! Input='"&crlf&\
   "'" & siold &"'") */\
			}/* error */\
		   else (*aChkI(oprpdl,1L,oprptr))= x;\
		   }
  /* ENDC */ /*  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

/* IFC EVAL(is_sail)=0 */
 /* THENC <<<FAILED!!!>>>

define vpush(x) = \Begin
		   If (valPtr_valPtr+1)>maxVALPDL
			Then
			Begin "error"
			isConstExprFlag_FALSE;
			Goto abortFOLDING;
# # # @V.52.49		ERR!HLR(
   "[FOLD!CONST!EXPR] VPUSH valPDL overflow! Input='"&CRLF&
   "'" & sIold &"'");
			End "error"
		   Else valPDL[valPtr]_(x);
		   End\;
  */ /* ENDC */

/* IFC EVAL(is_psail)=1 */
/* THENC */

#define vpush(x) {if ((++valptr)>30L)\
			{"error"\
			isconstexprflag= 0L;\
			goto abortfolding;\
/*  _2 _2 Ev.52.49		err_hlr(\
   "[FOLD!CONST!EXPR] VPUSH valPDL overflow! Input='"&crlf&\
   "'" & siold &"'") */\
			}/* error */\
		   else (*aChkF(valpdl,1L,valptr))= x;\
		   }
  /* ENDC */ /*  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#define opop (*aChkI(oprpdl,1L,(--oprptr)+1L))  /*  -- never get underflow! */
#define vpop (*aChkF(valpdl,1L,(--valptr)+1L))
#define oleftparen 0
#define orightparen 1
#define oplus 2
#define ominus 3
#define ounaryminus 4
#define otimes 5
#define odivide 6
#define omodulo 7
#define oland 8
#define oxor 9
#define olor 10
#define oand 11
#define oor 12
#define olnot 13
#define onot 14
#define olt 15
#define ogt 16
#define ole 17
#define oge 18
#define oeq 19
#define one 20
#define omax 21
#define omin 22
#define olsh 23
#define orsh 24
#define oquestion 25
#define ocolon 26
#define odelim 27

#define maxoprs 27L		/*  -- always the last one... */

 static INTEGER /*ARRAY*/ P_prec[13+(27-(-1)+1L)] = { 
   0,0,0,0,0,0,0,0,0,0,0,0,0, 1945, 	/*  -- oDELIM -1 */
   	1945, 	/*  -- oLEFTPAREN 0 */
   	0, 	/*  -- oRIGHTPAREN 1 */
   	6, 	/*  -- oPLUS 2 */
   	6, 	/*  -- oMINUS 3 */
   	1, 	/*  -- oUNARYMINUS 4 */
   	4, 	/*  -- oTIMES 5 */
   	4, 	/*  -- oDIVIDE 6 */
   	4, 	/*  -- oMODULO 7 */
   	9, 	/*  -- oLAND 8 */
   	10, 	/*  -- oXOR 9 */
   	11, 	/*  -- oLOR 10 */
   	13, 	/*  -- oAND 11 */
   	14, 	/*  -- oOR 12 */
   	3, 	/*  -- oLNOT 13 */
   	2, 	/*  -- oNOT 14 */
   	7, 	/*  -- oLT 15 */
   	7, 	/*  -- oGT 16 */
   	7, 	/*  -- oLE 17 */
   	7, 	/*  -- oGE 18 */
   	8, 	/*  -- oEQ 19 */
   	8, 	/*  -- oNE 19 */
   	12, 	/*  -- oMAX 21 */
   	12, 	/*  -- oMIN 22 */
   	5, 	/*  -- oLSH 23 */
   	5, 	/*  -- oRSH 24 */
   	15, 	/*  -- oQUESTION 25 */
   	15, 	/*  -- oCOLON 26 */
   	1945 };
 static INTEGER /*ARRAY*/ *prec  /* [-1:maxoprs] */ ;
     /*  -- operator precedence */

#define maxspclcsym 7			/*  -- special symbols test */
#define max2spclcsym 14L
 static STRING /*ARRAY*/ 
/*******************************************************************/
/* WARNING! Can't have OWN STRING declarations in non-MAIN module  */
/* or inner blocks! Suggest you move 'STRING spclcsym' declaration */
/* to global level in MAIN module and recompile.                   */
/*******************************************************************//*N.P.*/
/***************************/
/*<7>    ..                */
/*      .                  */
/*   /----------\          */
/*   | o         _______/  */
/*   |__         -------\  */
/*   \----------/          */
/*                         */
/*   A GROSS MOBY          */
/***************************/

P_spclcsym[26+(14-1+1L)] = { 
   0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, "lsh",	"<<",
   		/*  [1] */
   	"rsh",	">>",		/*  [2] */
   	"max",	"@u",		/*  [3] */
   	"min",	"@l",		/*  [4] */
   	"pnull", "0",		/*  [5] */
   	"true",	 "1",		/*  [6] */
   	"false", "0" };
 static STRING /*ARRAY*/ 
/*******************************************************************/
/* WARNING! Can't have OWN STRING declarations in non-MAIN module  */
/* or inner blocks! Suggest you move 'STRING spclcsym' declaration */
/* to global level in MAIN module and recompile.                   */
/*******************************************************************//*N.P.*/
/***************************/
/*<7>    ..                */
/*      .                  */
/*   /----------\          */
/*   | o         _______/  */
/*   |__         -------\  */
/*   \----------/          */
/*                         */
/*   A GROSS MOBY          */
/***************************/

*spclcsym  /* [1:max2spclcsym] */ ;
 /*  special oprs */

#define maxbkslashsym 10		/*  -- '\' symbols test */
 static STRING /*ARRAY*/ 
/*********************************************************************/
/* WARNING! Can't have OWN STRING declarations in non-MAIN module    */
/* or inner blocks! Suggest you move 'STRING bkslashsym' declaration */
/* to global level in MAIN module and recompile.                     */
/*********************************************************************//*N.P.*/
/***************************/
/*<7>    ..                */
/*      .                  */
/*   /----------\          */
/*   | o         _______/  */
/*   |__         -------\  */
/*   \----------/          */
/*                         */
/*   A GROSS MOBY          */
/***************************/

P_bkslashsym[26+(10-1+1L)] = { 
   0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, "r",	"015",
   	"n",	"012",
   	"\\",	"0134",
   	"\"",	"042",
   	"\'",	"047" };
 static STRING /*ARRAY*/ 
/*********************************************************************/
/* WARNING! Can't have OWN STRING declarations in non-MAIN module    */
/* or inner blocks! Suggest you move 'STRING bkslashsym' declaration */
/* to global level in MAIN module and recompile.                     */
/*********************************************************************//*N.P.*/
/***************************/
/*<7>    ..                */
/*      .                  */
/*   /----------\          */
/*   | o         _______/  */
/*   |__         -------\  */
/*   \----------/          */
/*                         */
/*   A GROSS MOBY          */
/***************************/

*bkslashsym  /* [1:maxbkslashsym] */ ;
/*  '\' operators */
		/*  -- Note: '\<left>' is replaced with '<right'> */

 static STRING /*ARRAY*/ 
/******************************************************************/
/* WARNING! Can't have OWN STRING declarations in non-MAIN module */
/* or inner blocks! Suggest you move 'STRING oprname' declaration */
/* to global level in MAIN module and recompile.                  */
/******************************************************************//*N.P.*/
/***************************/
/*<7>    ..                */
/*      .                  */
/*   /----------\          */
/*   | o         _______/  */
/*   |__         -------\  */
/*   \----------/          */
/*                         */
/*   A GROSS MOBY          */
/***************************/

P_oprname[26+(27-(-1)+1L)] = { 
   0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, "<NUL>",
   	"(", ")", "+", "-", "U-", "*", "/", "%", "&", "^", "|", 
   	"&&", "||", "~", "!", "<", ">", "<=", ">=", "==", "!=",
   	"@u", "@l", "<<", ">>", "?", ":", "#" };
 static STRING /*ARRAY*/ 
/******************************************************************/
/* WARNING! Can't have OWN STRING declarations in non-MAIN module */
/* or inner blocks! Suggest you move 'STRING oprname' declaration */
/* to global level in MAIN module and recompile.                  */
/******************************************************************//*N.P.*/
/***************************/
/*<7>    ..                */
/*      .                  */
/*   /----------\          */
/*   | o         _______/  */
/*   |__         -------\  */
/*   \----------/          */
/*                         */
/*   A GROSS MOBY          */
/***************************/

*oprname  /* [-1:maxoprs] */ ;
   /*   -- operator names  */

;

GC_EPOPS(dbug(02L,catlist(10L,GC_PUSHS("[FOLD!CONST!EXPR] useFoldingFlag="),
        pbool(usefoldingflag),GC_PUSHS(crlf),GC_PUSHS(" sI=\'"),
        GC_PUSHS(si),GC_PUSHS("\'"),GC_PUSHS(crlf),
        GC_PUSHS(" lastType=\'"),cvtypes(lasttype),GC_PUSHS("\'"))));

GC_POPS(si); /* from GC string PDL */
GC_MARKSSTKFRAME;   /* Mark G.C. STACK-FRAME */
GC_NOTESTRING(si); /* for G.C. */
GC_NOTESTRING(s); /* for G.C. */
GC_NOTESTRING(s1); /* for G.C. */
GC_NOTESTRING(s2); /* for G.C. */
GC_NOTESTRING(s3); /* for G.C. */
GC_NOTESTRING(so); /* for G.C. */
GC_NOTESTRING(a); /* for G.C. */
GC_NOTESTRING(siold); /* for G.C. */
GC_NOTESTRING(token); /* for G.C. */
GC_NOTESTRING(trash); /* for G.C. */
GC_NOTESTRING(lasttoken); /* for G.C. */
GC_NOTESTRING(nomoretext); /* for G.C. */
GC_NOTESTRING(lctoken); /* for G.C. */
GC_NOTESTRING(wsright); /* for G.C. */
GC_NOTEASTRING(spclcsym);	/* for G.C. */
GC_NOTEASTRING(bkslashsym);	/* for G.C. */
GC_NOTEASTRING(oprname);	/* for G.C. */
oprpdl = arymkI(1L,0L,15L,"oprpdl",'i');	/* Dynamic Array */
valpdl = arymkF(1L,0L,30L,"valpdl",'f');	/* Dynamic Array */
prec = aryHdrI(P_prec,1,-1,27,"prec",'i');	/* Dynamic Array */
spclcsym = aryHdrC(P_spclcsym,1,1,14,"spclcsym",'c');	/* Dynamic Array */
bkslashsym = aryHdrC(P_bkslashsym,1,1,10,"bkslashsym",'c');	/* Dynamic Array */
oprname = aryHdrC(P_oprname,1,-1,27,"oprname",'c');	/* Dynamic Array */
if (!usefoldingflag ||*si=='\0' ||isfragmentflag ||
   (((ps_itrash_= (lasttype&037))>0) ? (ps_itrash_+0140L)
				:0L)==codestring) 
	{/* premature exit - do nothing */
GC_EPOPS(dbug(02L,catlist(3L,GC_PUSHS("[FOLD!CONST!EXPR].0.0.0.1 No-fold sI=\'"),
        GC_PUSHS(si),GC_PUSHS("\'"))));
	;
	{aryfree(oprpdl,(INTEGER *)valpdl,
	        (INTEGER *)0);  /* Free List of Dynamic Arrays */
	GC_UNMARKSSTKFRAME;
	 GC_SRETURN(si);}
	}/* premature exit - do nothing */

pushnumber(lasttype);		/*  -- save it in case we fail */
oprptr= valptr= 0;		/*  -- initialize */
(*aChkI(oprpdl,1L,0L))= 27L;		/*  -- dummy terminator */
/*  # # @V.49.63 OPUSH(oDELIM) */
siold= si;	/*  -- Copy original string in case fail folding */

sawoctalflag= sawrealflag= 0L;
maxvals= 0;		/*  -- count tokens in expr to find simple expr */
didmacroexpandflag= 0L;
nomoretext= PNULL;
token= PNULL;
isconstexprflag= 1L;			/*  -- initially assume OK */
s= si;				/*  -- also make a copy */

i1= i= length(GC_PUSHS(si));		/*  -- so we don't mess around in the 
				beginning and if there is a single token
				we detect it */
while ((i>0) &&(ch= (*(si+i-1))) &&iswhitespace(ch)) --i;
if ((i>0) &&(i<i1))
	{/* extract trailing WS */
	wsright= subst(GC_PUSHS(si),i+1L,i1);		/*  -- save right WS for later */
	si= subst(GC_PUSHS(si),1L,i);			/*  -- shorten it */
	}/* extract trailing WS */
else wsright= PNULL;
/*  # # DBUG('2,"[FOLD!CONST!EXPR].0.0.0.2 i="&CVS(i)&" i1="&CVS(i1)&CRLF&" sI='"&sI&"'"&CRLF&" s='"&s&"'"&CRLF&" wsRight='"&wsRight&"'") */

do 	{/* scan and eval sI */
	while ((*si!='\0') &&iswhitespace((INTEGER)*si)) ps_itrash_= LOP(si);
	lasttoken= token;	/*  -- back up for UNARYMINUS */
/*  # # DBUG('2,"[FOLD!CONST!EXPR].0.0.1 lastToken='"&lastToken&"'"&CRLF&" sI='"&sI&"'") */

	/*  -- Get next C token */
	if (*si=='('&&equ(subst(GC_PUSHS(si),1L,9L),GC_PUSHS("(INTEGER)"))) si= subst(GC_PUSHS(si),10L,length(si));  /*  -- trash '(INTEGER)' cast */
	do {token= GC_EPOPS(get_c_token(&si,&nomoretext));} while (!((*si=='\0') ||
	        !(*token=='/'&&equ(subst(GC_PUSHS(token),1L,2L),GC_PUSHS("/*"))))); 
				/*  -- ignore comments... */
/*  # # DBUG('2,"[FOLD!CONST!EXPR].0.0.2 token='"&token&"'"&CRLF&" sI='"&sI&"'") */
	if (isdigit((INTEGER)*token) &&(*(token+length(token)-1))=='L') token= subst(GC_PUSHS(token),1L,length(token)-1L);	/*  -- trash trailing L
							cast */
/*  # # DBUG('2,"[FOLD!CONST!EXPR].0.0.3 token='"&token&"' oprPDL[oprPtr="&CVS(oprPtr)&"]="&CVS(oprPDL[oprPtr])&CRLF&" sI='"&sI&"'") */

	if ((*token==','&&!(omax<=(*aChkI(oprpdl,1L,oprptr)) && (*aChkI(oprpdl,1L,oprptr))<=24L)) ||
	    (*token==';')) 
		{/* abort on commas or <SC> */
		isconstexprflag= 0L;
/*  # # DBUG('2,"[FOLD!CONST!EXPR].0.0.4 isConstExprFlag="&PBOOL(isConstExprFlag)) */
		break;
		}/* abort on commas or <SC> */
	else if (*token==',') continue;	/*  -- ignore ','sif optimizing

	 */if (*token==*quote||*token==*singlequote)
		{/* map the 1st character to a number */
		/*  -- get the rest of the quoted string */
		ct= (INTEGER)*token;	/*  -- save quote char to match */
/*  # # @V.55.24	While (ch_LOP(sI)) Neq NULL Do
# # # @V.55.24		Begin "build quoted token"
# # # @V.55.24		token_token & ch */
/*  # # @V.55.24		If (ch="\")
# # # @V.55.24			Then token_token & LOP(sI)
# # # @V.55.24		Else If ch=cT
# # # @V.55.24			Then Done */
/*  # # @V.55.24		End "build quoted token" */
		if (*si=='\0'&&oprptr==0 &&valptr==0) isconstexprflag= 0L;  /*  -- Sorry,
						We do not optimize
						quoted stuff here If there
						is only one operand
						and no operators */
/*  # # DBUG('2,"[FOLD!CONST!EXPR].0.1 isConstExprFlag="&PBOOL(isConstExprFlag)&CRLF&" token='"&token&"'"&CRLF&" sI='"&sI&"'") */
		brk2= LOP(token);	/*  -- remove 1st quote */
		if (*token==('\\'/*WARNING*/))
			{/* is escape character */
			brk2= LOP(token);  /*  -- trash the '\'  */
			if (isdigit((INTEGER)*token))
				{/* get octal number */
				s1= "@";
				while ((*token!='\0') &&
				      isdigit((INTEGER)*token))
						s1= catlist(2L,GC_PUSHS(s1),GC_PUSHS(LOP(token)));
				brk2= LOP(s1);	/*  -- trash '@' */
				token= s1;
				}/* get octal number */
			else for (i=1;i<=10L;i+=2)
				if (*token==*((*aChkC(bkslashsym,1L,i)))) 
					{/* match */
					token= (*aChkC(bkslashsym,1L,i+1L));
					break;
					}/* match */
			}/* is escape character */
		else token= catlist(2L,GC_PUSHS("0"),cvos((brk2= LOP(token))));
/*  # # DBUG('2,"[FOLD!CONST!EXPR].0.1.1 token='"&token&"'") */
		}/* map the 1st character to a number */
	token= GC_EPOPS(cvlower(GC_PUSHS(token)));		/*  -- make sure!!! */

/*  # # DBUG('2,"[FOLD!CONST!EXPR].1 token='"&token&"' lastToken='"&lastToken&"'"&CRLF&" oprPDL["&CVS(oprPtr)&"]='"&oprName[oprPDL[oprPtr]]&"' valPDL["&CVS(valPtr)&"]='"&CVF(valPDL[valPtr])&"'"&CRLF&"  sI='"&sI&"'") */
/*  # # DBUG('2,"[FOLD!CONST!EXPR].1.1 token='"&token&"' maxVALS="&CVS(maxVALS)&" sI='"&sI&"'") */
	opr= -1;		/*  -- null operator */

	if ((*token!='\0') &&chk_id(GC_PUSHS(token)))
		{/* handle operand value */
		isoprflag= 0L;	/*  -- default to operand */
		ispclc= 0;		/*  -- index of match */
		lctoken= GC_EPOPS(cvlower(GC_PUSHS(token)));	/*  -- check as lower case */
		for (i=1;i<=14L;i+=2)
			{/* test special alphanums */
			s3= (*aChkC(spclcsym,1L,i));
			if (fast_equ(GC_PUSHS(lctoken),GC_PUSHS((*aChkC(spclcsym,1L,i)))))
				{/* match */
				ispclc= (i/2L)+1L;
				token= (*aChkC(spclcsym,1L,i+1L));  /*  -- map it */
/*  # # DBUG('2,"[FOLD!CONST!EXPR].1.1.1 iSpclC="&CVS(iSpclC)&" lcToken='"&lcToken&"' NEW token='"&token&"'"&CRLF&" sI='"&sI&"'") */
				break;
				}/* match */
			}/* test special alphanums */
		if ((ispclc==5) &&(oprptr==0&&valptr==0 &&*si=='\0')) isconstexprflag= FALSE	/*  -- keep PNULL intact
						     if it is the only
						     expr */; 
		else if (1<=ispclc && ispclc<=4)
			{/* map to special characters' */
/*  -- NOTE: These are binary function calls:
		LSH(bits,shiftCnt),
		RSH(bits,shiftCnt),
		MAX(a,b),
		MIN(a,b)
 */
			while ((*si!='\0') &&iswhitespace((INTEGER)*si)) ps_itrash_= LOP(si);
			if ((brk= LOP(si))!='(') isconstexprflag= 0L;  /*  -- trash
					the '(' and error if none */
/*  # # DBUG('2,"[FOLD!CONST!EXPR].1.1.1.1 sI='"&sI&"'") */
			isoprflag= 1L;
/*  # # @V.54.27		Case iSpclC Of
# # # @V.54.27		Begin "special operators"
# # # @V.54.27		"0"	 */	/*  -- no-op */
/*  # # @V.54.27		"1"	Begin "map 'lsh' to '<<'"
# # # @V.54.27		sI_"<" & sI */   /*  -- push it back */
/*  # # @V.54.27		sawOctalFlag_TRUE */
/*  # # @V.54.27		End "map 'lsh' to '<<'" */
/*  # # @V.54.27		"2"	Begin "map 'rsh' to '>>'"
# # # @V.54.27		sI_">" & sI */   /*  -- push it back */
/*  # # @V.54.27		sawOctalFlag_TRUE */
/*  # # @V.54.27		End "map 'rsh' to '>>'" */
/*  # # @V.54.27		"3"	Begin "map 'max' to '@u'"
# # # @V.54.27		sI_"u" & sI */   /*  -- push it back */
/*  # # @V.54.27		End "map 'max' to '@u'" */
/*  # # @V.54.27		"4"	Begin "map 'min' to '@l'"
# # # @V.54.27		sI_"l" & sI */   /*  -- push it back */
/*  # # @V.54.27		End "map 'min' to '@l'" */
/*  # # @V.54.27		End "special operators" */
/*  # # DBUG('2,"[FOLD!CONST!EXPR].1.1.1.2 sI='"&sI&"'") */
			}/* map to special characters' */

		else if (((idx= find_symbol(GC_PUSHS(token),&ct,&ht))>0))
			{/* try to expand macro constant */
			dw= (*aChkI(decldw,1L,idx));
/*  # # DBUG('2,"[FOLD!CONST!EXPR].1.2 token='"&token&"' idx="&CVS(idx)&CRLF&" dw=0"&CVOS(dw)&"='"&CVTYPES(dw)&"'") */
			if ((dw&02000L)==0||
			   (dw&040000L) ||
			   ((*aChkI(declnargs,1L,idx))>0)) 
				{/* failed */
				isconstexprflag= 0L;
/*  # # DBUG('2,"[FOLD!CONST!EXPR].1.2.1 token='"&token&"' isConstExprFlag="&PBOOL(isConstExprFlag)&CRLF&" dw=0"&CVOS(dw)&" ='"&CVTYPES(dw)&"'") */
				}/* failed */
			else	{/* Expand macro */
				pushnumber(midofexpr);
				midofexpr= 1L;  /*  -- in case quotes */
				pushnumber(usefoldingflag);
				usefoldingflag= 0L;
/*  # # DBUG('2,"[FOLD!CONST!EXPR].1.3 MACRO declSize[idx]='"&declSize[idx]&"'"&CRLF&PRT!NODE(hT)) */
				s1= (*aChkC(declsize,1L,idx));  /*  -- get macro body */
				if (dw&01000L) 
					{/* eval the sucker */
					s1= GC_EPOPS(cvevalquote(GC_PUSHS(s1))); /*  --
					NOTE: holds evaled macro */
/*  # # DBUG('2,"[FOLD!CONST!EXPR].1.3.1 CVEVAL'd s1='"&s1&"'"&CRLF&" old sI='"&sI&"'") */
					/*  -- PSAIL!EVAL s1 */
					midofexpr= 1L;  /*  -- in case 
							there is ".." */
					push(GC_PUSHS(oldtoken));
					oldtoken= "FOLD!CONST";
					s2= s1= GC_EPOPS(r_d_expr_map(GC_PUSHS(s1)));
					pop(&oldtoken);
					}/* eval the sucker */
				else s2= s1;  /*  -- is ALREADY evaled! */
/*  # # DBUG('2,"[FOLD!CONST!EXPR].1.3.2 s2='"&s2&"'") */
				popnumber(&usefoldingflag);
				popnumber(&midofexpr);
				lctoken= GC_EPOPS(get_c_token(&s2,&(s3= PNULL))); 
					/*  -- see if simple id or number */
				if (*s2!='\0') s1= catlist(3L,GC_PUSHS("("),GC_PUSHS(s1),GC_PUSHS(")")); /*  -- only
						add '()' if really
						need them */
				if (*si=='\0') si= s1; 
				else si= catlist(2L,GC_PUSHS(s1),GC_PUSHS(si));  /*  -- backup */
				didmacroexpandflag= 1L;
/*  # # DBUG('2,"[FOLD!CONST!EXPR].1.4 New sI='"&sI&"'") */
				/* NO-OP */;
				goto PS0002 /* scan and eval sI */;
				}/* Expand macro */
/*  # # DBUG('2,"[FOLD!CONST!EXPR].2 Expand token='"&token&"' isConstExprFlag="&PBOOL(isConstExprFlag)) */
			}/* try to expand macro constant */
		if (!isoprflag &&
		   !((sawrealflag= chk_real(GC_PUSHS(token))) ||
			chk_integer(GC_PUSHS(token)))) 
			{/* failed */
			isconstexprflag= 0L;
/*  # # DBUG('2,"[FOLD!CONST!EXPR].2.0.1 token='"&token&"' isConstExprFlag="&PBOOL(isConstExprFlag)&" isOPRflag="&PBOOL(isOPRflag)) */
			}/* failed */
		}/* handle operand value */
	else isoprflag= 1L;
/*  # # DBUG('2,"[FOLD!CONST!EXPR].2.1 token='"&token&"' isConstExprFlag="&PBOOL(isConstExprFlag)&" isOPRflag="&PBOOL(isOPRflag)&" sawOctalFlag="&PBOOL(sawOctalFlag)&" sawRealFlag="&PBOOL(sawRealFlag)&CRLF&" sI='"&sI&"'") */

	if (!isoprflag &&isdigit((INTEGER)*token))
		{/* push the OPERAND */
		STRING rtemp;
		INTEGER rbrk;
		GC_MARKSSTKFRAME;   /* Mark G.C. STACK-FRAME */
		GC_NOTESTRING(rtemp); /* for G.C. */
		if (sawrealflag) vnum= realscan(&rtemp= GC_EPOPS(makstr(cvt_e_to_at((INTEGER)*token))),&rbrk); 
		else if (*token=='0'&&((vnum= cvo(GC_PUSHS(token)))!=0))	{/* saw octal flag */
				sawoctalflag= 1L;
				}/* saw octal flag */
		else vnum= cvd(GC_PUSHS(token));  /*  -- NOTE: numbers
					are computed as numbers not
					string representations of
					numbers */
/*  # # DBUG('2,"[FOLD!CONST!EXPR].3 VPUSH vNum[token="&token&"]='"&CVG(vNum)&"'"&CRLF&" sawRealFlag="&PBOOL(sawRealFlag)&" sawOctalFlag="&PBOOL(sawOctalFlag)) */
		vpush(vnum);
		++maxvals;
		if (*si!='\0') {GC_UNMARKSSTKFRAME;
 goto PS0002 /* scan and eval sI */;};
		{GC_UNMARKSSTKFRAME;}}/* push the OPERAND */

	else if (*token!='\0')
		{/* handle Scanned OPERATOR */
/*  # # DBUG('2,"[FOLD!CONST!EXPR].3.1 token='"&token&"' lastToken='"&lastToken&"'"&CRLF&" oprPDL["&CVS(oprPtr)&"]='"&oprName[oprPDL[oprPtr]]&"' valPDL["&CVS(valPtr)&"]='"&CVF(valPDL[valPtr])&"'"&CRLF&"  sI='"&sI&"'"&"  isConstExprFlag="&PBOOL(isConstExprFlag)) */
/*  # # DBUG('2,"[FOLD!CONST!EXPR].3.1.0 token='"&token&"' lastToken='"&lastToken&"'"&" oprPtr="&CVS(oprPtr)&"valPtr="&CVS(valPtr)&CRLF&"  sI='"&sI&"'"&CRLF&"  isConstExprFlag="&PBOOL(isConstExprFlag)) */
		opr= -1;
		if (fast_equ(GC_PUSHS(token),GC_PUSHS("-")) &&
			((*lasttoken=='\0') ||(*lasttoken=='('))) opr= 
		4L; else for (i=0L;i<=26L;i++)
			{/* check for operator */
			s2= (*aChkC(oprname,1L,i));
			if (!fast_equ(GC_PUSHS(token),GC_PUSHS(s2))) continue;
			opr= i;
			break;
			}/* check for operator */

		if (opr==-1)
			{/* failed */
			isconstexprflag= 0L;
/*  # #  # # # DBUG('2,"[FOLD!CONST!EXPR].3.1.1 token='"&token&"' isConstExprFlag="&PBOOL(isConstExprFlag)) */
			break;
			}/* failed */

/*  # # DBUG('2,"[FOLD!CONST!EXPR].3.1.2 opr='"&CVS(opr)&"'='"&oprName[opr]&"' oprPDL["&CVS(oprPtr)&"]='"&oprName[oprPDL[oprPtr]]&"'"&CRLF&"  prec["&oprName[opr]&"]="&CVS(prec[opr])&" prec["&oprName[oprPDL[oprPtr]]&"]="&CVS(prec[oprPDL[oprPtr]])) */
		if ((opr!=-1) &&(opr!=1L) &&
		   (oprptr==0||
		    ((*aChkI(prec,1L,opr))<=(*aChkI(prec,1L,(*aChkI(oprpdl,1L,oprptr))))) ||
		    (opr==0L))) 
			{/* Push non ')' OPERATORs */
			opush(opr);
/*  # # DBUG('2,"[FOLD!CONST!EXPR].3.1.3 OPUSH opr="&CVS(opr)&"='"&oprName[opr]&"' sI='"&sI&"'") */
			goto PS0002 /* scan and eval sI */;
			}/* Push non ')' OPERATORs */
		}/* handle Scanned OPERATOR */

/*  # # DBUG('2,"[FOLD!CONST!EXPR].3.2 TEST-EVAL token='"&token&"' opr='"&oprName[opr]&"'"&CRLF&" oprPDL["&CVS(oprPtr)&"]='"&oprName[oprPDL[oprPtr]]&"' valPDL["&CVS(valPtr)&"]='"&CVF(valPDL[valPtr])&"'")
 *//*  # # DBUG('2,"[FOLD!CONST!EXPR].3.2.1  prec["&oprName[opr]&"]="&CVS(prec[opr])&" prec["&oprName[oprPDL[oprPtr]]&"]="&CVS(prec[oprPDL[oprPtr]])&CRLF&" sI='"&sI&"'"&" isConstExprFlag="&PBOOL(isConstExprFlag)) */

	if (isconstexprflag) 
	    {/* Start partial EVAL on OPR-PREC., EOF or ')' */
	    /*  -- NOTE: 0 is highest precedence and infinity lowest */
	    while (isconstexprflag &&(oprptr>0&&valptr>0) &&
		    (*si=='\0'||(opr==1L) ||
		     ((*aChkI(prec,1L,opr))>=(*aChkI(prec,1L,(*aChkI(oprpdl,1L,oprptr)))))))	{/* Eval the stacks */
			opr1= opop;    /*  -- get the operator<==oprPDL[] */
/*  # # DBUG('2,"[FOLD!CONST!EXPR].3.2.2 EVAL OPOP opr1="&CVS(opr1)&"='"&oprName[opr1]&"'"&CRLF&" oprPDL["&CVS(oprPtr)&"]='"&oprName[oprPDL[oprPtr]]&"' valPDL["&CVS(valPtr)&"]='"&CVF(valPDL[valPtr])&"'"&CRLF&" prec[opr1]="&CVS(prec[opr1])&" AND prec[oprPDL[oprPtr]]="&CVS(prec[oprPDL[oprPtr]])) */
			v1= 0.0;
			i2= v2= vpop;	/*  -- set up to save time */
			switch (opr1)
				{/* eval */
		/*[oLEFTPAREN]*/	case 0: {vpush(v2);
				goto PS0003 /* Eval the stacks */;; break;
				}
		/*[oRIGHTPAREN]*/	case 1: {/* starts OPR eval */; break; }
		/*[oPLUS]*/	case 2: vnum= v2+(v1= vpop); break;
		/*[oMINUS]*/	case 3: {v1= vpop; vnum= v1-v2;; break; }
		/*[oUNARYMINUS]*/	case 4: vnum= -v2; break;
		/*[oTIMES]*/	case 5: vnum= v2*(v1= vpop); break;
		/*[oDIVIDE]*/	case 6: {v1= vpop;
				vnum= (v2==0.0) ? 0.0 : v1/v2;; break;
				}
		/*[oMODULO]*/	case 7: {v1= vpop; vnum= Pfmod(v1,v2);; break; }
		/*[oLAND]*/	case 8: vnum= i2&(i1= (v1= vpop)); break;
		/*[oXOR]*/	case 9: vnum= i2^(i1= (v1= vpop)); break;
		/*[oLOR]*/	case 10: vnum= i2|(i1= (v1= vpop)); break;
		/*[oAND]*/	case 11: {v1= vpop;
				vnum= ((v2!=0.0) &&(v1!=0.0)) ? 
					 1L:0L;; break;
				}
		/*[oOR]*/		case 12: {v1= vpop;
				vnum= ((v2!=0.0) ||(v1!=0.0)) ? 
					 1L:0L;; break;
				}
		/*[oLNOT]*/	case 13: vnum= ~i2; break;
		/*[oNOT]*/	case 14: vnum= ((!i2)) ?   1L: break;case 15: default:  0; break;
		/*[oLT]*/	case 16: vnum= ((v2>(v1= vpop))) ?   1L: break;case 17: default:  0; break;
		/*[oGT]*/	case 18: vnum= ((v2<(v1= vpop))) ?   1L: break;case 19: default:  0; break;
		/*[oLE]*/	case 20: vnum= ((v2>=(v1= vpop))) ?   1L: break;case 21: default:  0; break;
		/*[oGE]*/	case 22: vnum= ((v2<=(v1= vpop))) ?   1L: break;case 23: default:  0; break;
		/*[oEQ]*/	case 24: vnum= ((v2==(v1= vpop))) ?   1L: break;case 25: default:  0; break;
		/*[oNE]*/	case 26: vnum= ((v2!=(v1= vpop))) ?   1L: break;case 27: default:  0; break;
		/*[oMAX]*/	case 28: vnum= MAX(v2,(v1= vpop)); break;
		/*[oMIN]*/	case 29: vnum= MIN(v2,(v1= vpop)); break;
		/*[oLSH]*/	case 30: {v1= i1= vpop;
					vnum= LSH(i1,i2);; break; }
		/*[oRSH]*/	case 31: {v1= i1= vpop;
					vnum= RSH(i1,i2);; break; }
		/*[oQUESTION]*/	case 32: {vpush(v2);
				isconstexprflag= 0L;  /*  -- error */; break;
				}
		/*[oCOLON]*/	case 33: {float be;
					INTEGER oldopr;
					v1= vpop; be= vpop;
					if ((oldopr= opop)!=25L) isconstexprflag= 0L;
					vnum= ((be) ?  v1 : v2);; break;
					}
		/*[oDELIM]*/	case 34: {vpush(v2);
				goto PS0003 /* Eval the stacks */;; break;
				}
				}/* eval */
/*  # # DBUG('2,"[FOLD!CONST!EXPR].3.3 opr1='"&oprName[opr1]&"'"&CRLF&"  (v1,v2)=("&CVF(v1)&","&CVF(v2)&") new vNum="&CVF(vNum)) */
			if (opr1!=0L&&opr1!=27L)
				{/* push operand */
				vpush(vnum);  /*  -- stack new operand */
/*  # # DBUG('2,"[FOLD!CONST!EXPR].3.3.1 VPUSH vNum[token="&token&"]='"&CVF(vNum)&"'") */
				}/* push operand */
			if ((valptr<1)) isconstexprflag= 0L; /*  -- BAD WFF */
			PS0003: ;}/* Eval the stacks */
	    if (isconstexprflag &&(opr!=-1) &&
	       (opr!=1L)) 
		{/* now push the operator with lower precedence */
		opush(opr);	/*  -- now push it */
/*  # # DBUG('2,"[FOLD!CONST!EXPR].3.3.2 OPUSH opr="&CVS(opr)&"='"&oprName[opr]&"'") */
		}/* now push the operator with lower precedence */
	    }/* Start partial EVAL on OPR-PREC., EOF or ')' */
 	PS0002: ;}/* scan and eval sI */
 while (!((*si=='\0') ||
	        !(*token=='/'&&equ(subst(GC_PUSHS(token),1L,2L),GC_PUSHS("/*")))));

/*  # # DBUG('2,"[FOLD!CONST!EXPR].4 oprPtr="&CVS(oprPtr)&" valPtr="&CVS(valPtr)&" maxVALS="&CVS(maxVALS)&CRLF&" isConstExprFlag="&PBOOL(isConstExprFlag)&" didMACROexpandFlag="&PBOOL(didMACROexpandFlag)&CRLF&" useIFCevalFlag="&PBOOL(useIFCevalFlag)) */
if (((!(oprptr==0&&valptr==1)) ||
   ((maxvals==1) &&!didmacroexpandflag)) &&
   (!useifcevalflag)) isconstexprflag= 0L;	/*  -- BAD WFF */

/*  --------------------------------------------------------- */
/*  . . . LAND HERE IF WE WERE EVER ABORTED. . .		    */
/*  --------------------------------------------------------- */
abortfolding:
/*  # # DBUG('2,"[FOLD!CONST!EXPR].4.1 isConstExprFlag="&PBOOL(isConstExprFlag)) */
popnumber(&lasttype);		/*  -- restore it in case failed */
if (!isconstexprflag) so= siold; 

else	{/* cvt computed value to string */
	if (valptr==1) v= vpop; 
	else v= 0.0;		/*  -- default value */
	getformat(&p,&q);
	setformat(0L,8L);	    /*  -- ok for DEC10 36-bit FP but .... */
/*  # # DBUG('2,"[FOLD!CONST!EXPR].4.1 v='"&CVF(v)&"'") */
	lasttype=lasttype&0777777774007L;
	if (((v-(i= v))!=0.0) ||
	   ((i1= (((ps_itrash_= (lasttype&037))>0) ? (ps_itrash_+0140L)
				:0L))==codereal) ||(i1==codedouble))	{/* cvt to REAL */
			pdw_ptype(lasttype,codereal);
			so= GC_EPOPS(makstr(cvt_at_to_e((INTEGER)*(GC_EPOPS(cvg(v))))));
/*  # # DBUG('2,"[FOLD!CONST!EXPR].4.1 sO='"&sO&"'") */
			so= GC_EPOPS(remove_white_ends(GC_PUSHS(so)));
			if (*so=='.') so= catlist(2L,GC_PUSHS("0"),GC_PUSHS(so));
			while ((*(so+length(so)-1))=='0')
				so= subst(GC_PUSHS(so),1L,length(so)-1L);
			}/* cvt to REAL */
	else	{/* cvt to integer */
		pdw_ptype(lasttype,codeinteger);
		if (sawoctalflag) so= catlist(2L,GC_PUSHS("0"),cvos(i)); else so= GC_EPOPS(cvs(i));
		so= GC_EPOPS(cvactual_to_expected(lasttype,lasttype,GC_PUSHS(so)));
					/*  -- cast to 32-bit integer */
		}/* cvt to integer */
	setformat(p,q);
	if (*wsright!='\0') so= catlist(2L,GC_PUSHS(wsright),GC_PUSHS(so));
	}/* cvt computed value to string */

GC_EPOPS(dbug(02L,catlist(23L,GC_PUSHS("[FOLD!CONST!EXPR].5 sO=\'"),GC_PUSHS(so),
        GC_PUSHS("\'"),GC_PUSHS(crlf),GC_PUSHS(" oprPDL["),cvs(
        oprptr),GC_PUSHS("]=\'"),
        GC_PUSHS(trash= GC_EPOPS(makstr((*aChkI(oprpdl,1L,
        oprptr))))),GC_PUSHS("\' valPDL["),cvs(valptr),
        GC_PUSHS("]=\'"),cvf((*aChkF(valpdl,1L,valptr))),
        GC_PUSHS("\'"),GC_PUSHS(crlf),GC_PUSHS("  sI=\'"),GC_PUSHS(
        si),GC_PUSHS("\'"),GC_PUSHS("  isConstExprFlag="),
        pbool(isconstexprflag),GC_PUSHS(crlf),GC_PUSHS(
        " lastType=\'"),cvtypes(lasttype),GC_PUSHS("\'"))));

;
{aryfree(oprpdl,(INTEGER *)valpdl,
        (INTEGER *)0);  /* Free List of Dynamic Arrays */
 GC_SRETURN(so);}
/*************************************************/
/*        U N U S E D    D E C L A R A T I O N S */
/* 'olsh'=<L0> macro integer                     */
/* 'omin'=<L0> macro integer                     */
/* 'one'=<L0> macro integer                      */
/* 'oeq'=<L0> macro integer                      */
/* 'oge'=<L0> macro integer                      */
/* 'ole'=<L0> macro integer                      */
/* 'ogt'=<L0> macro integer                      */
/* 'olt'=<L0> macro integer                      */
/* 'onot'=<L0> macro integer                     */
/* 'olnot'=<L0> macro integer                    */
/* 'oor'=<L0> macro integer                      */
/* 'oand'=<L0> macro integer                     */
/* 'olor'=<L0> macro integer                     */
/* 'oxor'=<L0> macro integer                     */
/* 'oland'=<L0> macro integer                    */
/* 'omodulo'=<L0> macro integer                  */
/* 'odivide'=<L0> macro integer                  */
/* 'otimes'=<L0> macro integer                   */
/* 'ominus'=<L0> macro integer                   */
/* 'oplus'=<L0> macro integer                    */
/* 'a'=<L0> string                               */
/* 'h'=<L0> integer                              */
/*************************************************/
/************************************************/
/*<2>                                           */
/*        \-\  -----*                           */
/*         \ \  \  /                            */
/*          \ \  \/                             */
/*           \ \ /                              */
/*            \ \                               */
/*             \-\                              */
/*  ROUGH WEATHER LIES AHEAD, GET SHIP SHAPE!   */
/************************************************/



#undef maxbkslashsym
#undef max2spclcsym
#undef maxspclcsym
#undef maxoprs
#undef odelim
#undef ocolon
#undef oquestion
#undef orsh
#undef olsh
#undef omin
#undef omax
#undef one
#undef oeq
#undef oge
#undef ole
#undef ogt
#undef olt
#undef onot
#undef olnot
#undef oor
#undef oand
#undef olor
#undef oxor
#undef oland
#undef omodulo
#undef odivide
#undef otimes
#undef ounaryminus
#undef ominus
#undef oplus
#undef orightparen
#undef oleftparen
#undef vpop
#undef opop
#undef vpush
#undef opush
#undef maxvalpdl
#undef maxoprpdl
}/* FOLD!CONST!EXPR */
