/* FILE: get_number.c
                            Peter Lemkin 

                   Image Processing Section, DCBD
                      National Cancer Institute
                        FCRF Bld 469, Rm 150B
                         Frederick, MD 21701

			    June 7, 1987

     COPYRIGHT NOTICE - Copyright (c) 1983-1988 by Peter F. Lemkin.
                        All rights reserved.
	Permission is granted to any individual or institution to
        copy, use or modify this program, except for explicitly
        commercial purposes, provided this copyright notice is
        included in its entirety in all derived works.  No warranty
        of the software or of the accuracy of any documentation
        surrounding it is expressed or implied, and the author does
        not acknowledge any liability resulting from program or
        documentation errors.

              Derived from PSAIL.SAI V.43.50   7-19-84
 */


/* File: pslex.c, being converted 12/25/87, 08:38:10PM*/
/* PSAIL compiler, version: June 9, 1987 V.57.10(OPTIMIZED) 	*/

#include "pshead.h"	/* header file to include other header files */


/*  ------------------------------------------------------- */
/*INTERNAL*/  STRING get_number(s)
	STRING *s;
/*  ------------------------------------------------------- */
{/* GET!NUMBER */
/*  --
.LABEL [<++PSAIL PROCEDURE: GET!NUMBER>]
.INDEX ++PSAIL PROCEDURE: GET!NUMBER
DESCRIPTION:
	Scan a number: Integer, fraction or scientific notation from
s and return it. The SAIL manual [Reiser] page 42 gives the BNF for
the number lexical scanner. 

	<nbr> ::= <sign><real nbr>
	<real nbr> ::= <decimal nbr> | <decimal nbr><exponent> |
			<exponent>
	<decimal nbr> ::= <integer> | <integer> . | 
			<integer> . <integer> | . <integer>
	<integer> ::= <digit> | <integer> <digit>
	<exponent> ::= <base10><sign><integer>
	<base10> := @ | E | e
	<sign> ::= + | - | <NULL>

	Also set the global variable 'lastType' to the primitive type
of INTEGER or REAL as a side effect.

NOTE: that you should check for a number before coming here since we
assume that there is some number in saiStatement.

SIDE-EFFECTS:

PROBLEMS/KNOWN-BUGS:

.ENDLABEL [<++PSAIL PROCEDURE: GET!NUMBER>]
 */
STRING sr;
INTEGER ch			/*  -- current character */,
	chold			/*  -- previous character */,
	chnext			/*  -- next character */,
	sawdp			/*  -- saw decimal point */,
	sawe			/*  -- saw E (exponent) character */,
	sawexpsign;		/*  -- saw <sign> of exponent */

;

GC_EPOPS(dbug(01000L,catlist(3L,GC_PUSHS("[GET!NUMBER] s=\'"),GC_PUSHS(*s),GC_PUSHS("\'"))));
GC_MARKSSTKFRAME;   /* Mark G.C. STACK-FRAME */
GC_NOTESTRING(s); /* for G.C. */
GC_NOTESTRING(sr); /* for G.C. */
sawdp= sawe= sawexpsign= 0L;
ch= chnext= (INTEGER)*(sr= PNULL);

while ((*(*s)!='\0') &&
	(isdigit((INTEGER)*(*s)) ||
	 ((*(*s)=='.'&&
	  ((chnext= (*(*s+2-1)))=='e'||chnext=='E' ||isdigit(chnext) ||
	   !isvariable(chnext)) &&
	  (!sawdp)) ||

	 ((*(*s)=='E'||*(*s)=='e') &&
	  (!sawe) &&
	  ( ((chnext= (*(*s+2-1)))=='+'||chnext=='-') &&
	     !sawexpsign) ||
	    isdigit(chnext))
	 )
	))	{/* eat it up */
/*  # # DBUG('1000,"[GET!NUMBER].1 sR='"&sR&"' s='"&s&"' sawE="&PBOOL(sawE)&" sawDP="&PBOOL(sawDP)&" sawExpSign="&PBOOL(sawExpSign)) */
		chold= ch;		/*  -- backup previous character */
		ch= LOP(*s);
		if ((ch=='e')) ch= 'E';	/*  -- Use caps */
		sr= catlist(2L,GC_PUSHS(sr),makstr(ch));
		chnext= (INTEGER)*(*s);		/*  -- get new 1st character */
/*  # # DBUG('1000,"[GET!NUMBER].1.1 chNext='"&(sTemp_chNext)&"' ISDIGIT(chNext)="&PBOOL(ISDIGIT(chNext))&" ISVARIABLE(chNext)="&PBOOL(ISVARIABLE(chNext))) */
		if ((ch=='E')) 
			{/* handle exponent */
			sawe= 1L;
			if ((chnext=='+') ||(chnext=='-')) 
				{/* saw sign */
				sawexpsign= 1L;
				sr= catlist(2L,GC_PUSHS(sr),GC_PUSHS(LOP(*s)));
				}/* saw sign */
			}/* handle exponent */
		else if ((ch=='.')) sawdp= 1L;
		}/* eat it up */

/* NO-OP */;

lasttype= 0L;		/*  -- only has primitive type attribute */
if (sawdp ||sawe) pdw_ptype(lasttype,codereal); 
	else pdw_ptype(lasttype,codeinteger);
GC_EPOPS(dbug(01000L,catlist(5L,GC_PUSHS("[GET!NUMBER].2 sR=\'"),GC_PUSHS(sr),
        GC_PUSHS("\' lastType=\'"),cvtypes(lasttype),GC_PUSHS("\'"))));
;
GC_SRETURN(sr)
}/* GET!NUMBER */
