/* FILE: hash_c.c
                            Peter Lemkin 

                   Image Processing Section, DCBD
                      National Cancer Institute
                        FCRF Bld 469, Rm 150B
                         Frederick, MD 21701

			    June 7, 1987

     COPYRIGHT NOTICE - Copyright (c) 1983-1988 by Peter F. Lemkin.
                        All rights reserved.
	Permission is granted to any individual or institution to
        copy, use or modify this program, except for explicitly
        commercial purposes, provided this copyright notice is
        included in its entirety in all derived works.  No warranty
        of the software or of the accuracy of any documentation
        surrounding it is expressed or implied, and the author does
        not acknowledge any liability resulting from program or
        documentation errors.

              Derived from PSAIL.SAI V.43.50   7-19-84
 */


/* File: pssymt.c, being converted 12/25/87, 02:03:55PM*/
/* PSAIL compiler, version: June 9, 1987 V.57.10(OPTIMIZED) 	*/

#include "pshead.h"	/* header file to include other header files */


/*  ------------------------------------------------------- */
/*INTERNAL*/  INTEGER hash_c(sym)
	STRING sym;
/*  ------------------------------------------------------- */
{/* HASH!C */
/*   --
.LABEL [<++PSAIL PROCEDURE: HASH!C>]
.INDEX ++PSAIL PROCEDURE: HASH!C
DESCRIPTION:
	Lookup existing hash entry pointer (1 to maxSymbols-1) for
String 'str' if exists else return 0. Set 'lastHfree' to first free
symtab hole to be used later by PUT!HASH!C(). Note a hole in the
symtab is indicated by either a NULL or a BELL ('07) symbol name. The
BELL is stuffed in the table by procedure REM!SYMBOLS() when all
global and local symbols with the same name are trashed. The
'lastHcnt' is the number of probes required for this symbol. The
'hDeleted' is the FIRST '07 type slot found in the search and will be
used to set 'lastHfree' after the search.

@V.54.85 - based on DO!WATCH PROFILE statistics optimize this function
(its called a LOT which is really not surprising). @V.55.20 added two
errStatistics[] counters: errNHASH!C and errNPROBES!HASH!C to let us
compute the mean probes/access. If it is high then we can diddle the
hashing algorithm.

We could reduce the number of collisions by using a double hash
algorithm discussed in Standish,T.,"Data Structure Techniques",
Addison-Wesley, 1980 page 148. Instead of recomputing the probe by
adding the prime number to it add primeNbr[key] so that all collisions
don't follow the same path. A fast way to do this is to have a ready
table of prime numbers and then (key Land '7) to pick out say 1 out
of 8 of them (or '17 etc). The following table lists statistics for
several different simple overflow hashing functions without secondary
hashing.

PSAIL-VERSION	.SAI-file   #probes/#calls = P/C    %full  Hash fct
-------------   ---------   --------------   ----   -----  --------
@V.56.01        PSPPRT.SAI   99284/54501     1.82   80%    Knott (h)
@V.55.20        DEMO.SAI    138046/75337     1.83   86%    Knott (h)
@V.56.01        PSCVT.SAI   355809/93479     3.80   95%    Knott (h)
@V.56.15        DEMO.SAI    119280/75588     1.57   86%    Knuth (h+h)
@V.56.38        DEMO.SAI    122259/75780     1.61   86%    Knott (h+h+h)

SIDE-EFFECTS:

PROBLEMS/KNOWN-BUGS:

.ENDLABEL [<++PSAIL PROCEDURE: HASH!C>]
 */
INTEGER h0, 
	h, 
	ch, 
	hdeleted, 
	docvlwrflag;
STRING s, 
	s1, 
	smatch;
/*  # # PRELOAD!WITH
	3, 5, 7, 11, 13, 17, 19, 23 */   /*  -- or any primes???? */
/*  # # SAFE Integer Array h2Prime[0:7] */

;

GC_POPS(sym); /* from GC string PDL */
GC_MARKSSTKFRAME;   /* Mark G.C. STACK-FRAME */
GC_NOTESTRING(sym); /* for G.C. */
GC_NOTESTRING(s); /* for G.C. */
GC_NOTESTRING(s1); /* for G.C. */
GC_NOTESTRING(smatch); /* for G.C. */
docvlwrflag= hdeleted= lasthfree= lasthcnt= h0= 0;
if (*sym=='\0') { {GC_UNMARKSSTKFRAME;
  GC_NRETURN(0);} }
smatch= s= sym;			/*  -- Make copy of string to eat up */
while ((*s!='\0')) 
	{/* build hash number */
	ch= LOP(s);
	if (('A'<=ch && ch<='Z')) ch= (docvlwrflag= ch)-0242L;    /*  -- only lower
							 case */
/*  -- NOTE that we set doCVLWRflag above to non-null if any UPPER CASE
char is seen in the string. Then later we must spend the effort
converting the string to lower case since that is what we store in
the symbol table.
 */

/*  --	The following is the current hashing function used. NOTE:
 *only one should be active!!!!
 */
/*  # #	h0_h0 + ch */		/*  -- Knott's 'MLAB' hashing fct used
				 prior to PSAIL version @V.56.15 */
	h0= (h0+h0) +ch;	/*  -- Knuth's 'TANGLE' program hashing
				fct used after PSAIL version @V.56.15.
				NOTE: we only use this to reduce the
				avg # of probes. If it does not then
				go back to Knott's less expensive
				function */
/*  # # 	h0_(h0 + h0 + h0) + ch */	/* @V.56.38 suggested by 
                                                  Gary Knott.
				Next try (after lowercase it)
			 		h0_h0 + (LOP(s) LSH 8) + LOP(s).
				    */
	}/* build hash number */

if (docvlwrflag) 
	{/* cvt to lower here rather than CVLOWER() for speed */
	s= sym;
	smatch= "@";
	while ((ch= LOP(s))!=0)
		smatch= catlist(2L,GC_PUSHS(smatch),
        GC_PUSHS(('A'<=ch && ch<='Z') ? GC_EPOPS(makstr((ch-0242L) ))
                                      : ch));
	ch= LOP(smatch);			/*  -- trash '@' */
	}/* cvt to lower here rather than CVLOWER() for speed */

h= h0= MAX((h0 %1699L),1L);		/*  -- Never start at zero! */

/*  # # DBUG('1,"[HASH!C] sym="&sym&" h0="&CVS(h0)&" pcntHfull="&CVS(100.0*nHsymbols/maxSymbols)) */

while ((h!=0)) 
	{/* Hash it */
	lasthfree= h;			/*  -- Global used by PUT!HASH */
	if (*((s1= (*aChkC(saisym,1L,h))))=='\0')
		{/* opps -- end of the line */
		h= 0;		/*  -- does not exist for RETRIEVAL only */
		goto PS0001 /* Hash it */;
		}/* opps -- end of the line */
	else if (fast_equ(GC_PUSHS(smatch),GC_PUSHS(s1))) break  /* Hash it */; 		/*  -- found it
	 */if ((*s1==007) &&(hdeleted!=0)) hdeleted= h;	/*  -- save 1st deleted entry */
	if ((++lasthcnt)>=1699L)
		{/* Something smells - gross failure! */
		err_hlr(catlist(4L,GC_PUSHS("[HASH!C] DRYROT!! HASH!C("),GC_PUSHS(sym),
        GC_PUSHS(")] table overflow failure. lastHcnt="),cvs(
        lasthcnt)));
		goto PS0001 /* Hash it */;
		}/* Something smells - gross failure! */
	h= ((h+h0) %1699L); 	/*  -- compute next hash probe */
	}/* Hash it */
PS0001: 

if (hdeleted!=0) lasthfree= hdeleted;   /*  -- for possible use by PUT!HASH!C */

/*  # # DBUG('1,"[HASH!C].1 sym="&sym&" h="&CVS(h)&" lastHfree="&CVS(lastHfree)&" lastHcnt="&CVS(lastHcnt)) */


/*  # # If (lastHfree=0 And lastHcnt<maxSymbols)
	Then ERR!HLR("[HASH!C] DRYROT! HASH!C(" & sym &
			")]=0 with table NOT full!"&crlf&
			" lastHcnt=" &CVS(lastHcnt) &
			" lastHfree=" &CVS(lastHfree) &
			" h=" &CVS(h)) */
/*  # # If Not (0 < h < maxSymbols)
	Then DBUG('1,"[HASH!C].2 ILLEGAL sym='"&sym&"' h="&CVS(h)&" LENGTH(sym)="&CVS(LENGTH(sym))) */

{if (usereportflag) (*aChkI(errstatistics,1L,34L))= (*aChkI(errstatistics,1L,34L))+lasthcnt; }	/*  -- total # of probes */
{/* INCSTAT */ if (usereportflag) (*aChkI(errstatistics,1L,33L))= (*aChkI(errstatistics,1L,33L))+1L;
		     }/* INCSTAT */		/*  -- total # of HASH!C calls */
;
GC_NRETURN(h)
}/* HASH!C */
