/* FILE: map_quotes.c
                            Peter Lemkin 

                   Image Processing Section, DCBD
                      National Cancer Institute
                        FCRF Bld 469, Rm 150B
                         Frederick, MD 21701

			    June 7, 1987

     COPYRIGHT NOTICE - Copyright (c) 1983-1988 by Peter F. Lemkin.
                        All rights reserved.
	Permission is granted to any individual or institution to
        copy, use or modify this program, except for explicitly
        commercial purposes, provided this copyright notice is
        included in its entirety in all derived works.  No warranty
        of the software or of the accuracy of any documentation
        surrounding it is expressed or implied, and the author does
        not acknowledge any liability resulting from program or
        documentation errors.

              Derived from PSAIL.SAI V.43.50   7-19-84
 */


/* File: psparse.c, being converted 12/25/87, 09:34:41PM*/
/* PSAIL compiler, version: June 9, 1987 V.57.10(OPTIMIZED) 	*/

#include "pshead.h"	/* header file to include other header files */


/*  ----------------------------------------------------- */
/*INTERNAL*/  STRING map_quotes(si)
	STRING *si;
/*  ----------------------------------------------------- */
{/* MAP!QUOTES */
/*   --
.LABEL [<++PSAIL PROCEDURE: MAP!QUOTES>]
.INDEX ++PSAIL PROCEDURE: MAP!QUOTES
DESCRIPTION:
	Copy a quoted string from 'sI' where the first " is removed
prior to this call.

	This code is NOT redundent with the code in
GET!NEXT!STATEMENT mapper since it handles the \" and \' mappings if
the isCOPY!TOflag is FALSE which was set by COPY!EITHER. The switch
may also be set by whoever wants to scan a complex quoted string
without doing a C eval.

	Quoted string conversion rules:
		SAIL		C	Data type
		----		-	----------
		"\"	==>	'\\'	CODEinteger
		"'"	==>	'\''	CODEinteger
		""	==>	'\0'	CODEinteger
		""""	==>	'\"'	CODEinteger
		"a"	==>	'a'	CODEinteger
		"abc" 	==>	"abc"	CODEstring
		"a\b" 	==>	"a\\b"	CODEstring
		"ab'c" 	==>	"ab\'c"	CODEstring
		"a""c" 	==>	"a\"c"	CODEstring
		"""abc"""   ==>	"\"abc\"" CODEstring
NOTE: if NOT doing COPY!EITHER (i.e. NOT isCOPY!TOflag) then do
this conversion, otherwise NO conversion!

SIDE-EFFECTS:

PROBLEMS/KNOWN-BUGS:

.ENDLABEL [<++PSAIL PROCEDURE: MAP!QUOTES>]
 */
STRING s, 
	trash, 
	s1, 
	s2, 
	ws;
INTEGER ch, 
	isquote, 
	quotecnt, 
	atype, 
	docvtflag, 
	isnpconstflag;

;

GC_EPOPS(dbug(02000L,catlist(8L,GC_PUSHS("[MAP!QUOTES] |sI|="),cvs(length(GC_PUSHS(*si))),
        GC_PUSHS(" sI=\'"),GC_PUSHS(*si),GC_PUSHS("\'"),GC_PUSHS(
        crlf),GC_PUSHS(" isCOPY!TOflag="),pbool(iscopy_toflag))));

GC_MARKSSTKFRAME;   /* Mark G.C. STACK-FRAME */
GC_NOTESTRING(si); /* for G.C. */
GC_NOTESTRING(s); /* for G.C. */
GC_NOTESTRING(trash); /* for G.C. */
GC_NOTESTRING(s1); /* for G.C. */
GC_NOTESTRING(s2); /* for G.C. */
GC_NOTESTRING(ws); /* for G.C. */
s= quote;			/*  -- set first quote */
quotecnt= 1;			/*  -- set balanced " counter to 1st " */
isquote= 1L;			/*  -- Set in-quote-flag */
isnpconstflag= 0L;		/*  -- set N.P. string const-flag */
ws= PNULL;			/*  -- no leading white space */
atype= 0;			/*  -- type to be returned */
pdw_ptype(atype,codestring);

while ((*(*si)!='\0') &&isquote)
	{/* copy character */
	s1= GC_EPOPS(scan(si,22L,&ch));	/*  -- break on ", ' and \  */
	if (*s1!='\0') s= catlist(2L,GC_PUSHS(s),GC_PUSHS(s1));  /*  -- except quotes */
	if (ch==*singlequote&&!iscopy_toflag) s= catlist(2L,GC_PUSHS(s),GC_PUSHS("\\\'")); else if (ch==('\\'/*WARNING*/)&&!iscopy_toflag) s= catlist(2L,GC_PUSHS(s),GC_PUSHS("\\\\")); else if (ch==*quote)
		{/* found another quote */
		++quotecnt;
		if ((*(*si)!=*quote))
			{/* see if finished */
			if (((quotecnt&01)==0)) isquote= 0L;
			}/* see if finished */

		else 	{/* get double quote */
			++quotecnt;
			ch= LOP(*si); /*trash*/
			if (iscopy_toflag) s= catlist(2L,GC_PUSHS(s),GC_PUSHS(quote)); else if (!ismacroflag &&
				((*(!s+length(!s)-1))==('\\'/*WARNING*/))) s= catlist(2L,GC_PUSHS(s),GC_PUSHS("\\"));
			}/* get double quote */
		s= catlist(2L,GC_PUSHS(s),GC_PUSHS(quote)); /*always 2nd quote*/
		}/* found another quote */

	else if (!iscopy_toflag &&(0<ch && ch<40))
		{/* Special C code mapping for magic characters */
/*  # # DBUG('2000,"[MAP!QUOTES].0.1 ch='"&(ps!sTrash!_ch)&"'"&CRLF&" s='"&s&"'"&CRLF&" sI='"&sI&"'") */
/*  -- @V.53.22 NOTE: There is a portability problem here. Most C
compilers will NOT compile a multiline string constant with the CRLF
or \n intact in the string! So we must fake it out and handle it in
<sairun.c>'s printf() function. The problem is that will the C compiler
	1) convert \15\12 to the actual inline single 2 character CRLF
or	2) leave it as the 6 character \15\12 to be interpreted by
	   printf().
For now, we will emit the "\15\12" & CRLF and
note in the PSUSER.MEM the need to be careful....
 */
/*  # # @V.53.20	s_s & (If (ch=CR And sI=LF)
# # # @V.53.20			Then ("\n\" & CR & LOP(sI))
# # # @V.53.20		  Else "\" & (If ch=FF
# # # @V.53.20					Then "f"
# # # @V.53.20			       Else If ch=TAB
# # # @V.53.20					Then "t"
# # # @V.53.20			       Else If ch=CR
# # # @V.53.20					Then "r"
# # # @V.53.20			       Else If ch='7
# # # @V.53.20					Then "a"
# # # @V.53.20			       Else CVOS(ch)) ) */
		if ((ch==*cr&&*(*si)==*lf)) {/* CRLF */
			     s= catlist(4L,GC_PUSHS(s),GC_PUSHS("\\15\\12\\"),GC_PUSHS(cr),GC_PUSHS(
        LOP(*si)));
			     isnpconstflag= 1L;
			     }/* CRLF */
		else s= catlist(2L,GC_PUSHS(s),makstr(ch));
		}/* Special C code mapping for magic characters */

	else if (ch!='\0') s= catlist(2L,GC_PUSHS(s),makstr(ch)); 		/*  -- build quoted string

 *//*  # # DBUG('200,"[MAP!QUOTES].1 ch='"&(ps!sTrash!_ch)&"'"&CRLF&" |s|="&CVS(LENGTH(s))&" s='"&s&"' |sI|="&CVS(LENGTH(sI))&" sI='"&sI&"'") */
	}/* copy character */
/*  # # DBUG('200,"[MAP!QUOTES].2 |s|="&CVS(LENGTH(s))&" s='"&s&"' |sI|="&CVS(LENGTH(sI))&" sI='"&sI&"'") */

	/*  -- Copy White space and Test if map "x" to 'x'    */
if (!iscopy_toflag) 
	{/* check if map to '.' */
	if (iswhitespace((INTEGER)*saistatement)) map_whitespace(&ws);	/*  -- trailing WS */
	docvtflag= ((((ps_itrash_= (lasttype&037))>0) ? (ps_itrash_+0140L)
				:0L)!=codestring) &&(*(*si)!='&');
	if (docvtflag)
	   {/* do CVT flag */
	   if ((*s==*quote) &&(length(GC_PUSHS(s))==3) &&
	      ((*(s+2-1))!=*quote) &&((*(s+3-1))==*quote)) 
		{/* optimize quote'quote to '' */
		pdw_ptype(atype,codeinteger);
		if ((*(s+2-1))==('\''/*WARNING*/)) s= "\'\\\'\'"; 
		else s= catlist(3L,GC_PUSHS("\'"),makstr(*(s+2-1)),GC_PUSHS("\'"));
		isquote= 0L;
		}/* optimize quote'quote to '' */
	   else if (fast_equ(GC_PUSHS(s),GC_PUSHS(twoquotes))) 
		{/* optimize quote quote to '\0' */
		pdw_ptype(atype,codeinteger);
		s= "\'\\0\'";    		/*  -- ok, do what C likes... */
		isquote= 0L;
		}/* optimize quote quote to '\0' */
	   }/* do CVT flag */
	else if (fast_equ(GC_PUSHS(s),GC_PUSHS(fourquotes))) 
		{/* optimize 4 quotes to '\quote' */
		s= catlist(3L,GC_PUSHS("\'\\"),GC_PUSHS(quote),GC_PUSHS("\'"));   	/*  -- i.e '\"'  */
		isquote= 0L;
		pdw_ptype(atype,codeinteger);
		}/* optimize 4 quotes to '\quote' */
	}/* check if map to '.' */

if (*ws!='\0') s= catlist(2L,GC_PUSHS(s),GC_PUSHS(ws));		/*  -- add White space on at the end */
if (!iscopy_toflag) lasttype= atype;		/*  -- set the type */

if (isnpconstflag) err_n_p(&s);
GC_EPOPS(dbug(0200L,catlist(9L,GC_PUSHS("[MAP!QUOTES].3 |s|="),cvs(length(GC_PUSHS(s))),
        GC_PUSHS(" s=\'"),GC_PUSHS(s),GC_PUSHS("\' |sI|="),cvs(
        length(GC_PUSHS(*si))),GC_PUSHS(" sI=\'"),GC_PUSHS(*si),
        GC_PUSHS("\'"))));
/* NO-OP */;
/* NO-OP */;
;
GC_SRETURN(s)
/*************************************************/
/*        U N U S E D    D E C L A R A T I O N S */
/* 's2'=<L0> string                              */
/* 'trash'=<L0> string                           */
/*************************************************/

}/* MAP!QUOTES */
